# Copyright (C) 2024 Tobias Bode
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

import time
import sys

import jax
from jax import lax, config
import jax.numpy as jnp
import flax

from autopdex import assembler, seeder, geometry, solver, solution_structures, plotter, utility, models, spaces

config.update("jax_enable_x64", True)
config.update("jax_compilation_cache_dir", './cache')


### Transport equation as moving least square solution structure with least square variational method


### Definition of geometry and boundary conditions
# Linear transport problem

# Positive smooth distance functions that select the specific boundary segments
# Just inital condition
psdf_left = lambda x: geometry.psdf_trimmed_line(x, jnp.stack([0.,0.]), jnp.stack([0.,1.]))
psdf_bottom = lambda x: geometry.psdf_trimmed_line(x, jnp.stack([0.,0.]), jnp.stack([1.,0.]))
psdfs = [psdf_bottom, psdf_left]
psdf = lambda x: geometry.psdf_unification(x, psdfs)

# Positive smooth distance funcion of full domain (used for seeding nodes and integration points)
psdf_rect = lambda x: geometry.psdf_parallelogram(x, jnp.stack([0.,0.]), jnp.stack([0.,1.]), jnp.stack([1.,0.]))
psdf_domain = psdf_rect

# Boundary condition functions
initial_condition = lambda x: jnp.sin(2 * jnp.pi * x[0])
homogeneous = lambda x: 0.
b_cond = lambda x: geometry.transfinite_interpolation(x, psdfs, [initial_condition, homogeneous])

### Initialization and generation of nodes and integration points
# Regular node placement
nod_spacing = 0.02
(x_nodes, n_nodes, region_size) = seeder.regular_in_psdf(psdf_domain, [0.,0.], [1.,1.], nod_spacing)

# Gauss integration on background grid
int_spacing = nod_spacing
(x_int, w_int, n_int) = seeder.gauss_points_in_psdf(psdf_domain, [0.,0.], [1.,1.], int_spacing, order=2)

# Support radius for moving least square ansatz
support_radius = 3.01 * nod_spacing

# Definition of pde
pde = models.transport_equation(c = 0.5)

# ### Visualizer for boundary functions
# fun = lambda x: b_cond(x)
# plotter.pv_function_plot(x_int, fun)
# sys.exit()


### Neighborhood search
(num_neighbors, max_neighbors, min_neighbors, neighbor_list) = utility.search_neighborhood(x_nodes, x_int, support_radius)


### Settings
n_fields = 1
dofs0 = jnp.zeros((n_nodes, n_fields))
static_settings = flax.core.FrozenDict({
  'solution space': ('mls',),
  'shape function mode': 'compiled',
  'number of fields': (n_fields,),
  'assembling mode': ('sparse',),
  'maximal number of neighbors': (max_neighbors,),
  'order of basis functions': (2,),
  'weight function type': ('gaussian',),
  'variational scheme': ('least square pde loss',),
  'model': (pde,),
  'solution structure': ('dirichlet',),
  'boundary conditions': (b_cond,),
  'psdf': (psdf,),
  'solver type': 'linear',
  'solver backend': 'pardiso',
  'solver': 'lu',
  'verbose': 1,
  'connectivity': (utility.jnp_to_tuple(neighbor_list),),
})
settings = {
  'beta': (3,),
  'node coordinates': x_nodes,
  'integration coordinates': (x_int,),
  'integration weights': (w_int,),
  'support radius': (support_radius,),
}


# Precompute shape functions
settings = solution_structures.precompile(dofs0, settings, static_settings)


### Call solver
dofs, _ = solver.solver(dofs0, settings, static_settings, tol=1e-8)


### Preperation of postprocessing data
local_dofs = dofs[neighbor_list]
post_fun_vj = jax.jit(jax.vmap(solution_structures.solution_structure, (0,0,0,None,None,None)), static_argnames=['static_settings', 'set'])
data = post_fun_vj(x_int, jnp.arange(n_int), local_dofs, settings, static_settings, 0)

# Plotting with pyvista
plotter.pv_plot(x_int, data[:,0], export_vtk=False, show=True)



# Copyright (C) 2024 Tobias Bode
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

import time
import sys

import jax
from jax import lax, config
import jax.numpy as jnp
import flax
import equinox as eqx

from autopdex import seeder, geometry, solver, solution_structures, plotter, models, spaces

config.update("jax_enable_x64", True)
config.update("jax_compilation_cache_dir", './cache')



### Example: Dirichlet problem of the Laplace equation in 2d: [0,1]x[0,1] with cutout circle (midpoint=[0.5,0.5], radius=0.25)
# Neural network as user-defined solution space, boundary conditions through R functions



# Positive smooth distance functions that select the specific boundary segments
psdf_bottom = lambda x: geometry.psdf_trimmed_line(x, [0.,0.], [1.,0.])
psdf_right = lambda x: geometry.psdf_trimmed_line(x, [1.,0.], [1.,1.])
psdf_top = lambda x: geometry.psdf_trimmed_line(x, [1.,1.], [0.,1.])
psdf_left = lambda x: geometry.psdf_trimmed_line(x, [0.,1.], [0.,0.])
psdf_cutout = lambda x: geometry.psdf_nd_sphere_cutout(x, [0.5,0.5], 0.25)
psdf_rect = lambda x: geometry.psdf_parallelogram(x, [0.,0.], [0.,1.], [1.,0.])
psdfs = [psdf_bottom, psdf_right, psdf_top, psdf_left, psdf_cutout]

# Positive smooth distance funcion of full domain (used for seeding nodes and integration points)
psdf_domain = lambda x: geometry.r_equivalence(psdf_rect(x), psdf_cutout(x))

# Boundary condition functions
def homogeneous(x): # Bottom, right and top boundaries
  return 0
def bc_0(x): # Left boundary
  return jax.lax.sin(jnp.pi * x[0])

# Assignment of boundary conditions
dirichlet_conditions = [bc_0, homogeneous, homogeneous, homogeneous, lambda x: 1.0]

# Transfinite interpolation of boundary segments
b_cond = lambda x: geometry.transfinite_interpolation(x, psdfs, dirichlet_conditions)
psdf = psdf_domain

# Integration points
int_spacing = 0.02
(x_int, n_int, region_size) = seeder.regular_in_psdf(psdf_domain, [0.,0.], [1.,1.], int_spacing)
w_int = (region_size / n_int) * jnp.ones(n_int)

# Definition of pde
pde = models.poisson(source_fun=None)


### User defined solution space
key = jax.random.PRNGKey(123)
class CNN(eqx.Module):
    layers: list

    def __init__(self, key):
        key1, key2, key3 = jax.random.split(key, 3)
        # Neural network with n_dim inputs and n_fields outputs
        self.layers = [
            eqx.nn.Linear(2, 20, key=key1),
            jax.nn.silu,
            eqx.nn.Linear(20, 20, key=key2),
            jax.nn.silu,
            eqx.nn.Linear(20, 1, key=key3),
            jax.nn.silu,
            ]

    def __call__(self, x):
        for layer in self.layers:
            x = layer(x)
        return x

key, subkey = jax.random.split(key, 2)
model = CNN(subkey)

dofs0, static = eqx.partition(model, eqx.is_array)

# Due to the structure of dofs, in this way it will only work for assembler.integrate_functional and the wrapped solvers 'lbfgs', 'bfgs', 'gradient descent' and 'nonlinear cg' (not for assembler.assemble_residual) 
def neural_network(x, int_point_number, dofs, settings, static_settings, set):
  parametrized_network = eqx.combine(dofs, static)
  return parametrized_network(x)


### Settings
n_fields = 1
static_settings = flax.core.FrozenDict({
  'solution space': ('user',),
  'user solution space function': (neural_network,),
  'shape function mode': 'direct',
  'number of fields': (n_fields,),
  'assembling mode': ('dense',),
  'variational scheme': ('least square pde loss',),
  'model': (pde,),
  'solution structure': ('dirichlet',),
  'boundary conditions': (b_cond,),
  'psdf': (psdf,),
  'solver type': 'minimize',
  'solver': 'lbfgs',
  'verbose': 1,
})
settings = {
  'integration coordinates': (x_int,),
  'integration weights': (w_int,),
}


### Here, the built-in wrapper for jaxopt minimizers is used. One may built a user-defined minimizer based on the function assembler.integrate_functional(dofs, settings, static_settings)
start = time.time()
dofs, _ = solver.solver(dofs0, settings, static_settings, tol=1e-12, maxiter=1000)
print("Solver time: ", time.time() - start)


### Preperation of postprocessing data, --> unoptimized solution structure fulfills boundary conditions, but not pde
post_fun_vj = jax.jit(jax.vmap(solution_structures.solution_structure, (0,0,None,None,None,None)), static_argnames=['static_settings', 'set'])
data = post_fun_vj(x_int, jnp.arange(n_int), dofs0, settings, static_settings, 0)
plotter.pv_plot(x_int, data[:,0], export_vtk=False, show=True)


### Preperation of postprocessing data, --> solution structure fulfills boundary conditions, and pde approximately
post_fun_vj = jax.jit(jax.vmap(solution_structures.solution_structure, (0,0,None,None,None,None)), static_argnames=['static_settings', 'set'])
data = post_fun_vj(x_int, jnp.arange(n_int), dofs, settings, static_settings, 0)
plotter.pv_plot(x_int, data[:,0], export_vtk=False, show=True)


### Preperation of postprocessing data, --> solution space does neighter fulfill boundary conditions nor pde
post_fun_vj = jax.jit(jax.vmap(spaces.solution_space, (0,0,None,None,None,None)), static_argnames=['static_settings', 'set'])
data = post_fun_vj(x_int, jnp.arange(n_int), dofs, settings, static_settings, 0)
plotter.pv_plot(x_int, data[:,0], export_vtk=False, show=True)



# Copyright (C) 2024 Tobias Bode
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.


def test_different_solvers():
    import sys
    import time

    import jax
    from jax import config
    import jax.numpy as jnp
    import flax

    from autopdex import seeder, geometry, solver, solution_structures, plotter, utility, models, assembler

    config.update("jax_enable_x64", True)

    ### Dirichlet problem of transient heat conduction equation (moving least square solution structure)
    # Comparison of different linear solvers


    # Positive smooth distance functions that select the specific boundary segment
    # Time dimension is always the last dimension (here the second)
    psdf_bottom = lambda x: geometry.psdf_trimmed_line(x, jnp.stack([0.,0.]), jnp.stack([1.,0.]))
    psdf_right = lambda x: geometry.psdf_trimmed_line(x, jnp.stack([1.,0.]), jnp.stack([1.,1.]))
    psdf_left = lambda x: geometry.psdf_trimmed_line(x, jnp.stack([0.,1.]), jnp.stack([0.,0.]))
    psdfs = [psdf_bottom, psdf_right, psdf_left]

    # Positive smooth distance funcion of full domain (used for seeding nodes and integration points)
    psdf_rect = lambda x: geometry.psdf_parallelogram(x, jnp.stack([0.,0.]), jnp.stack([0.,1.]), jnp.stack([1.,0.]))
    psdf_domain = psdf_rect

    # Initial condition
    def bc_bottom(x):
        return jax.lax.sin(jnp.pi * x[0])

    # Dirichlet boundary conditions on left an right side
    def bc_left(x):
        return 0.
    def bc_right(x):
        return 0.

    # Assignment of boundary conditions (has to be in the same order as the psdfs)
    dirichlet_conditions = [bc_bottom, bc_right, bc_left]

    # Transfinite interpolation of boundary segments
    b_cond = lambda x: geometry.transfinite_interpolation(x, psdfs, dirichlet_conditions)
    psdf = lambda x: geometry.psdf_unification(x, psdfs)

    ### Initialization and generation of nodes and integration points
    nod_spacing = 0.2
    (x_nodes, n_nodes, region_size) = seeder.regular_in_psdf(psdf_domain, [0.,0.], [1.,1.], nod_spacing)

    # Integration points
    int_spacing = nod_spacing
    (x_int, w_int, n_int) = seeder.gauss_points_in_psdf(psdf_domain, [0.,0.], [1.,1.], int_spacing, order=2)

    # Support radius for moving least square ansatz
    support_radius = 2.01 * nod_spacing

    # Definition of pde
    diffusivity = lambda x: 0.5
    pde = models.heat_equation(diffusivity_fun=diffusivity)

    ### Neighborhood search
    (num_neighbors, max_neighbors, min_neighbors, neighbor_list) = utility.search_neighborhood(x_nodes, x_int, support_radius)


    ### Settings
    n_fields = 1
    dofs0 = jnp.zeros((n_nodes, n_fields))
    static_settings = flax.core.FrozenDict({
    'solution space': ('mls',),
    'shape function mode': 'compiled',
    'number of fields': (n_fields,),
    'assembling mode': ('sparse',),
    'maximal number of neighbors': (max_neighbors,),
    'order of basis functions': (1,),
    'weight function type': ('gaussian',),
    'variational scheme': ('least square pde loss',),
    'model': (pde,),
    'solution structure': ('dirichlet',),
    'boundary conditions': (b_cond,),
    'psdf': (psdf,),

    # See different solver options below.
    # Remarks:
    # For small to medium problems on cpu, pardiso is a good choice
    # If tangent can not be loaded in RAM: try matrix-free solvers
    # For very large problems: try multigrid preconditioners on (gpu) cluster (petsc, pyamg)
    # For itterative solvers: setting tol or maxiter to low can reduce the solution accuracy

    # 'solver type': 'linear',
    # 'solver backend': 'pyamg',                # for further options see pyamg documentation
    # 'solver': 'cg',                           # cg, bcgs
    # 'type of preconditioner': 'ruge stuben',  # ruge stuben, smoothed aggregation 

    # 'solver type': 'linear',
    # 'solver backend': 'petsc',        # available options see PETSc documentation
    # 'solver': 'bcgs',
    # 'type of preconditioner': 'ilu',

    # 'solver type': 'linear',
    # 'solver backend': 'jax',              # if possible, uses matrix-free algorithms
    # 'solver': 'cg',                       #cg, bicgstab, gmres(, cg normal, lu, cholesky, qr, jacobi)
    # 'type of preconditioner': 'jacobi',   # 'jacobi', 'none'
    # 'hvp type': 'fwdrev',                 # fwdrev, revrev (only for symmetric matrices), linearize, assemble

    # 'solver type': 'linear',
    # 'solver backend': 'scipy',  # only on cpu
    # 'solver': 'lapack',         # lapack, umfpack

    'solver type': 'linear',
    'solver backend': 'pardiso', # only on cpu
    'solver': 'lu', # lu, qr

    # # Only works for problems where a functional can be minimized (e.g. least square pde loss, least square function approximation)
    # 'solver type': 'minimize',  # maxiter and tol may have to be adjusted (see effect on optimized L2 error); 
    # 'solver': 'lbfgs',          # lbfgs, gradient descent, nonlinear cg, bfgs, gauss newton, levenberg marquart

    'verbose': -1,
    'connectivity': (utility.jnp_to_tuple(neighbor_list),),
    })
    settings = {
    'beta': (3,),
    'node coordinates': x_nodes,
    'integration coordinates': (x_int,),
    'integration weights': (w_int,),
    'support radius': (support_radius,),
    }


    ### Precompute shape functions and pass them to e.g. static_settings...
    settings = solution_structures.precompile(dofs0, settings, static_settings)

    ### Call solver
    # dofs, _ = solver.solver(dofs0, settings, static_settings, tol=1e-8, maxiter=1000)
    # assert jnp.allclose(dofs, jnp.asarray([[ 0.45187177], [-0.41965862], [-0.513136  ], [-0.38273308], [-0.8074755 ], 
    #                                        [-1.40044741], [-1.18542015], [-0.8464102 ], [-0.8074755 ], [-1.40044741], 
    #                                        [-1.18542015], [-0.8464102 ], [ 0.45187177], [-0.41965862], [-0.513136  ], 
    #                                        [-0.38273308]])), 'Problem with pardiso solver'

    static_settings = static_settings.copy({
    'solver backend': 'scipy',
    'solver': 'lapack',
    })

    dofs, _ = solver.solver(dofs0, settings, static_settings, tol=1e-8, maxiter=1000)
    assert jnp.allclose(dofs, jnp.asarray([[ 0.45187177], [-0.41965862], [-0.513136  ], [-0.38273308], [-0.8074755 ], 
                                           [-1.40044741], [-1.18542015], [-0.8464102 ], [-0.8074755 ], [-1.40044741], 
                                           [-1.18542015], [-0.8464102 ], [ 0.45187177], [-0.41965862], [-0.513136  ], 
                                           [-0.38273308]])), 'Problem with scipy solver'

    # static_settings = static_settings.copy({
    # 'solver backend': 'pyamg',
    # 'solver': 'cg',
    # 'type of preconditioner': 'ruge stuben',
    # })

    # dofs, _ = solver.solver(dofs0, settings, static_settings, tol=1e-8, maxiter=1000)
    # assert jnp.allclose(dofs, jnp.asarray([[ 0.45187177], [-0.41965862], [-0.513136  ], [-0.38273308], [-0.8074755 ], 
    #                                        [-1.40044741], [-1.18542015], [-0.8464102 ], [-0.8074755 ], [-1.40044741], 
    #                                        [-1.18542015], [-0.8464102 ], [ 0.45187177], [-0.41965862], [-0.513136  ], 
    #                                        [-0.38273308]])), 'Problem with pyamg solver'

    # static_settings = static_settings.copy({
    # 'solver backend': 'petsc',
    # 'solver': 'bcgs',
    # 'type of preconditioner': 'ilu',
    # })

    # dofs, _ = solver.solver(dofs0, settings, static_settings, tol=1e-8, maxiter=1000)
    # assert jnp.allclose(dofs, jnp.asarray([[ 0.45187177], [-0.41965862], [-0.513136  ], [-0.38273308], [-0.8074755 ], 
    #                                        [-1.40044741], [-1.18542015], [-0.8464102 ], [-0.8074755 ], [-1.40044741], 
    #                                        [-1.18542015], [-0.8464102 ], [ 0.45187177], [-0.41965862], [-0.513136  ], 
    #                                        [-0.38273308]])), 'Problem with petsc solver'

    static_settings = static_settings.copy({
    'solver backend': 'jax', 
    'solver': 'cg', 
    'type of preconditioner': 'jacobi',
    'hvp type': 'fwdrev',
    })

    dofs, _ = solver.solver(dofs0, settings, static_settings, tol=1e-8, maxiter=1000)
    assert jnp.allclose(dofs, jnp.asarray([[ 0.45187177], [-0.41965862], [-0.513136  ], [-0.38273308], [-0.8074755 ], 
                                           [-1.40044741], [-1.18542015], [-0.8464102 ], [-0.8074755 ], [-1.40044741], 
                                           [-1.18542015], [-0.8464102 ], [ 0.45187177], [-0.41965862], [-0.513136  ], 
                                           [-0.38273308]])), 'Problem with jax solver'


test_different_solvers()


# solution_structures.py
# Copyright (C) 2024 Tobias Bode
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

"""
In solving boundary or initial value problems, one generally seeks a function
that satisfies both certain boundary conditions and differential equations.
In solution structures, the boundary conditions (or at least part of them)
are already incorporated into the solution space. This provides greater flexibility
in the choice of variational methods and, depending on the approach, allows for
exact satisfaction of the boundary conditions.

This module provides functions to compute the solution structure at a given
integration point using different modes, as well as to precompile and evaluate discrete
solution structures by evaluating and storing shape functions, boundary conditions,
and projection operators. Precompilation can improve the efficiency,
but one has to consider how many discrete derivatives need to be evaluated for the given solution structure and pde.

The idea of the RFM (Rvachev-function method) solution structures is basically to multiply the solution space :math:`\phi`
with a smooth function :math:`\omega` that vanishes on the boundary and to add subsequently a function
:math:`b` that satisfies the boundary conditions: :math:`\\tilde{\phi} = \omega \phi + b`.
The initial approximation :math:`\phi` can e.g. be a finite element ansatz, but also some approach in which boundary conditions
can not simply nodally imposed, such as smooth meshfree shape functions or a neural network.
The global weighting function :math:`\omega` can be constructed for example based on Rvachev functions 
that can be used as smooth distance functions to certain primitive elements. The boundary function :math:`b` can be set up
via transfinite interpolation in case there are different conditions on different segments of the boundary. 
The above procedure works well e.g. for a Laplace equation with Dirichlet conditions, but special techniques have to be used for 
more complex problems, see e.g. `Rvachev and Sheiko (1995) <https://doi.org/10.1115/1.3005099>`_. These kind of solution structures 
may be implemented as a user-specific solution structure 'user' (see the source code for examples). 

Here, a general technique is implemented with the solution structure type 'first order set' which can handle 
arbitrary order PDEs with multiple coupled boundary conditions. 
However, the prerequisite is that the problem is reformulated into a set of first-order problems and that a suitable variational 
method is used, such as a norm-equivalent first-order least square formulation 
(see, e.g., `Gunzburger and Bochev (2009) <https://link.springer.com/book/10.1007/b13382>`_). 
This means that not only, for example, the temperature field :math:`\\theta` serves as a primary variable, but also its derivatives, such as
:math:`\\theta_{,X}` and :math:`\\theta_{,t}`. If we gather all the primary fields into the vector :math:`\\boldsymbol{\\phi}`, 
boundary conditions of the form :math:`\\boldsymbol{a}_i \cdot \\boldsymbol{\\phi} = b_i` can be imposed by the following approach:
:math:`\\tilde{\\boldsymbol{\\phi}} = \\boldsymbol{P} \\cdot \\boldsymbol{\\phi} + \\tilde{\\boldsymbol{b}}`
with the projector field :math:`\\boldsymbol{P} = \\boldsymbol{I} - \\sum_i (1 - \omega_i) \\boldsymbol{a}_i \\otimes \\boldsymbol{a}_i`,
where :math:`\\boldsymbol{I}` is the identity matrix, :math:`\omega_i` is a boundary segment selector as a smooth distance funcion and the 
set of :math:`\\boldsymbol{a}_i` have to be orthogonal to each other at the boundary segments with each having a length of 1. 
The boundary function can be set up based on transfinite interpolations based on 
:math:`\\tilde{\\boldsymbol{b}} = \sum_i b_i \\boldsymbol{a}_i`. An example can be found e.g. in 
`space_time_fos_dirichlet_neumann_robin.py <https://github.com/BodeTobias/AutoPDEx/tree/main/examples/heat_conduction/space_time_fos_dirichlet_neumann_robin.py>`_
"""


import jax
import jax.numpy as jnp

from autopdex import spaces
from autopdex.utility import jit_with_docstring


@jit_with_docstring(static_argnames=['static_settings', 'set'])
def solution_structure(x, int_point_number, local_dofs, settings, static_settings, set):
  """
  Computes the solution structure at a given integration point using various predefined schemes.

  Args:
    x (jnp.ndarray): Coordinates of the integration point.
    int_point_number (int): Integration point number.
    local_dofs (jnp.ndarray): Local degrees of freedom.
    settings (dict): Settings for the computation.
    static_settings (dict): Static settings for the computation.
    set (str): Identifier for the current set.

  Returns:
    array: Computed solution structure.
  
  Notes:
    This function supports different solution structures, including 'dirichlet', 'first order set',
    'second order set', 'nodal imposition', and 'off'. Depending on the shape function mode ('direct'
    or 'compiled'), it evaluates shape functions and computes the solution structure accordingly.
    
    For 'dirichlet', 'first order set', and 'second order set' structures, boundary conditions and 
    projections are applied to the space. The function uses either direct evaluation or precomputed 
    shape functions and their derivatives based on the shape function mode.
  """
  try:
    structure_type = static_settings['solution structure'][set]
  except KeyError:
    print('The solution structure was not set. Continuing with \'off\'... ')
    structure_type = 'off'

  try:
    shp_fun_mode = static_settings['shape function mode']
  except KeyError:
    shp_fun_mode = None

  if structure_type == 'dirichlet':###Todo: obsolet?
    # Direct evaluation of shape functions
    if shp_fun_mode == 'direct':
      space = spaces.solution_space(x, int_point_number, local_dofs, settings, static_settings, set)
      b_condition = static_settings['boundary conditions'][set](x)
      psdf_d = static_settings['psdf'][set](x)
      return psdf_d * space + b_condition

    # Evaluation with precomputed shape functions -> shape function mode
    elif shp_fun_mode == 'compiled':
      (f, dx, ddx) = settings['compiled shape functions'][set]
      (bc, dbc, ddbc) = settings['compiled bc'][set]
      (psdf, dpsdf, ddpsdf) = settings['compiled projection'][set]
      shps = compiled_function_2(x, f[int_point_number], dx[int_point_number], ddx[int_point_number])
      space = (shps @ local_dofs)
      bc_x= _compiled_function_2(x, bc[int_point_number], dbc[int_point_number], ddbc[int_point_number])
      psdf_x = _compiled_function_2(x, psdf[int_point_number], dpsdf[int_point_number], ddpsdf[int_point_number])
      return psdf_x * space + bc_x
    else:
      assert False, 'Shape function mode neither specified as \'direct\' nor \'compiled\'!'

  elif structure_type == 'first order set' or structure_type == 'second order set':
    # alpha has to be normalized at boundaries
    # consider only orthogonal alphas
    # Direct evaluation of shape functions
    if shp_fun_mode == 'direct':
      space = spaces.solution_space(x, int_point_number, local_dofs, settings, static_settings, set)
      phi_0 = static_settings['boundary conditions'][set](x)
      alpha = static_settings['boundary coefficients'][set](x)
      omega = static_settings['psdf'][set](x)
      n_fields = static_settings['number of fields'][set]

      # Boundary function
      phi = phi_0 @ alpha

      # Projections deletes the influence of space where omega_i are zero in the direction of alpha_i
      def project_i(omega_i, alpha_i):
        return (1 - omega_i) * jnp.outer(alpha_i, alpha_i)
      projection = jnp.identity(n_fields) - jax.vmap(project_i, (0,0), 0)(omega, alpha).sum(axis=0)

      structure = projection @ space + phi
      return structure

    # Evaluation with precomputed shape functions -> shape function mode
    elif shp_fun_mode == 'compiled':
      if structure_type == 'first order set':
        (f, dx) = settings['compiled shape functions'][set]
        (bc, dbc) = settings['compiled bc'][set]
        (proj, dproj) = settings['compiled projection'][set]

        shps = compiled_function_1(x, f[int_point_number], dx[int_point_number])
        bc_x = _compiled_function_1(x, bc[int_point_number], dbc[int_point_number])
        proj_x = _compiled_function_1(x, proj[int_point_number], dproj[int_point_number])
      elif structure_type == 'second order set':
        (f, dx, ddx) = settings['compiled shape functions'][set]
        (bc, dbc, ddbc) = settings['compiled bc'][set]
        (proj, dproj, ddproj) = settings['compiled projection'][set]

        shps = compiled_function_2(x, f[int_point_number], dx[int_point_number], ddx[int_point_number])
        bc_x = _compiled_function_2(x, bc[int_point_number], dbc[int_point_number], ddbc[int_point_number])
        proj_x = _compiled_function_2(x, proj[int_point_number], dproj[int_point_number], ddproj[int_point_number])

      space = (shps @ local_dofs)
      structure = proj_x @ space + bc_x
      return structure
    else:
      assert False, 'Shape function mode neither specified as \'direct\' nor \'compiled\'!'
  
  elif structure_type == 'nodal imposition' or structure_type == 'off':
     # Just forward the space
    if shp_fun_mode == 'compiled':
      (f, dx) = settings['compiled shape functions'][set]
      shps = compiled_function_1(x, f[int_point_number], dx[int_point_number])
      space = (shps @ local_dofs)
      return space
    else:
      return spaces.solution_space(x, int_point_number, local_dofs, settings, static_settings, set)
    
  elif structure_type == 'user':
    return static_settings['user solution structure'][set](x, int_point_number, local_dofs, settings, static_settings, set)
  else:
    assert False, 'Solution structure type not available.'

@jit_with_docstring(static_argnames=['static_settings'])
def precompile(dofs0, settings, static_settings, max_diff_order=1):
  """
  Manually precompiles solution structures by evaluating and storing shape functions, boundary conditions, 
  and projection operators for different integration points and domains.

  Args:
    dofs0 (jnp.ndarray): Initial degrees of freedom.
    settings (dict): Settings for the computation.
    static_settings (dict): Static settings for the computation.
    max_diff_order (int, optional): Maximum order of derivatives to precompute. Default is 1.

  Returns:
    dict: Updated settings with precomputed shape functions, boundary conditions, and projections.

  Notes:
    This function loops over all sets of integration points or domains and compiles the necessary 
    components for different solution structures. The supported solution structures include:
    
    - 'first order set': Compiles shape functions, boundary conditions, and projection operators with first derivatives.
    - 'second order set': Compiles shape functions, boundary conditions, and projection operators with second derivatives.
    - 'dirichlet': Compiles shape functions, boundary conditions, and projection operators with second derivatives.
    - 'nodal imposition' or 'off': Compiles shape functions with derivatives up to max_diff_order.
    
    The precomputed components are stored in the settings dictionary under 'compiled shape functions', 
    'compiled bc', and 'compiled projection'.

    Warning: The compiled shape function mode does currently not support implicit differentiation, 
    since only the derivatives with respect to the coordinates (and not other quantities in the settings) are set.
  """
  # Loop over all sets of integration points/ domains
  num_sets = len(static_settings['assembling mode'])
  compiled_shps = ()
  compiled_bcs = ()
  compiled_proj = ()
  for set in range(num_sets):

    # Compilation of different terms in solution structures 
    mode = static_settings['solution structure'][set]

    if mode == 'first order set':
      b_cond = static_settings['boundary conditions'][set]
      psdf = static_settings['psdf'][set]
      x_int = settings['integration coordinates'][set]
      b_coeff = static_settings['boundary coefficients'][set]

      compiled_shps += (spaces.precompute_shape_functions(dofs0, settings, static_settings, set, num_diff=1),)
      compiled_bcs += (precompute_coupled_boundary(b_cond, b_coeff, x_int),)
      compiled_proj += (precompute_projection(psdf, b_coeff, x_int),)

    elif mode == 'nodal imposition' or mode == 'off':
      compiled_shps += (spaces.precompute_shape_functions(dofs0, settings, static_settings, set, num_diff=max_diff_order),)
      compiled_bcs += (None,)
      compiled_proj += (None,)

    elif mode == 'second order set':
      b_cond = static_settings['boundary conditions'][set]
      psdf = static_settings['psdf'][set]
      x_int = settings['integration coordinates'][set]
      b_coeff = static_settings['boundary coefficients'][set]

      compiled_shps += (spaces.precompute_shape_functions(dofs0, settings, static_settings, set, num_diff=2),)
      compiled_bcs += (precompute_coupled_boundary(b_cond, b_coeff, x_int, num_diff=2),)
      compiled_proj += (precompute_projection(psdf, b_coeff, x_int, num_diff=2),)

    elif mode == 'dirichlet':
      b_cond = static_settings['boundary conditions'][set]
      psdf = static_settings['psdf'][set]
      x_int = settings['integration coordinates'][set]

      compiled_shps += (spaces.precompute_shape_functions(dofs0, settings, static_settings, set, num_diff=2),)
      compiled_bcs += (precompute_function_and_derivatives(b_cond, x_int, num_diff=2),)
      compiled_proj += (precompute_function_and_derivatives(psdf, x_int, num_diff=2),)
   
    else: 
      assert False, 'There is no default compilation defined for the specified solution structure!'

  settings['compiled shape functions'] = compiled_shps
  settings['compiled bc'] = compiled_bcs
  settings['compiled projection'] = compiled_proj

  jax.debug.print('Compilation of solution structure finished.')
  return settings

@jit_with_docstring(static_argnames=['fun','num_diff'])
def precompute_function_and_derivatives(fun, x_int, num_diff=1):
  """
  Precomputes a function and its derivatives up to a specified order at given integration points.

  Args:
    fun (callable): The function to precompute.
    x_int (jnp.ndarray): Integration points.
    num_diff (int, optional): Number of derivatives to compute (0, 1, or 2). Default is 1.

  Returns:
    tuple: Precomputed function values and derivatives.
      - If num_diff == 0: (fun_i,)
      - If num_diff == 1: (fun_i, dfun_i)
      - If num_diff == 2: (fun_i, dfun_i, ddfun_i)
  """
  fun_i = jax.jit(jax.vmap(fun, (0), 0))(x_int)
  if num_diff == 0:
    return (fun_i)

  dfun_i = jax.jit(jax.vmap(jax.jacrev(fun), (0), 0))(x_int)
  if num_diff == 1:
    return (fun_i, dfun_i)

  ddfun_i = jax.jit(jax.vmap(jax.jacfwd(jax.jacrev(fun)), (0), 0))(x_int)
  if num_diff == 2:
    return (fun_i, dfun_i, ddfun_i)

  assert False, 'Number of differentiations not implemented!'

@jit_with_docstring(static_argnames=['phi_0_fun','alpha_fun','num_diff'])
def precompute_coupled_boundary(phi_0_fun, alpha_fun, x_int, num_diff=1):
  """
  Precomputes coupled boundary conditions and their derivatives up to a specified order at given integration points.

  Args:
    phi_0_fun (callable): Boundary condition function.
    alpha_fun (callable): Coefficient function.
    x_int (jnp.ndarray): Integration points.
    num_diff (int, optional): Number of derivatives to compute (0, 1, or 2). Default is 1.

  Returns:
    tuple: Precomputed boundary condition values and derivatives.
      - If num_diff == 0: (fun_i,)
      - If num_diff == 1: (fun_i, dfun_i)
      - If num_diff == 2: (fun_i, dfun_i, ddfun_i)
  """
  def phi(x):
    phi_0 = phi_0_fun(x)
    alpha = alpha_fun(x)
    return phi_0 @ alpha

  fun = phi
  fun_i = jax.jit(jax.vmap(fun, (0), 0))(x_int)
  if num_diff == 0:
    return (fun_i)
  
  dfun_i = jax.jit(jax.vmap(jax.jacrev(fun), (0), 0))(x_int)
  if num_diff == 1:
    return (fun_i, dfun_i)

  ddfun_i = jax.jit(jax.vmap(jax.jacfwd(jax.jacrev(fun)), (0), 0))(x_int)
  if num_diff == 2:
    return (fun_i, dfun_i, ddfun_i)

@jit_with_docstring(static_argnames=['psdf_fun','alpha_fun','num_diff'])
def precompute_projection(psdf_fun, alpha_fun, x_int, num_diff=1):
  """
  Precomputes the projection operator and its derivatives up to a specified order at given integration points.

  Args:
    psdf_fun (callable): Projection function.
    alpha_fun (callable): Coefficient function.
    x_int (jnp.ndarray): Integration points.
    num_diff (int, optional): Number of derivatives to compute (0, 1, or 2). Default is 1.

  Returns:
    tuple: Precomputed projection values and derivatives.
      - If num_diff == 0: (fun_i,)
      - If num_diff == 1: (fun_i, dfun_i)
      - If num_diff == 2: (fun_i, dfun_i, ddfun_i)
  """
  def projection(x):
    alpha = alpha_fun(x)
    omega = psdf_fun(x)
    n_fields = alpha.shape[-1]

    def project_i(omega_i, alpha_i):
      return (1 - omega_i) * jnp.outer(alpha_i, alpha_i)
    return jnp.identity(n_fields) - jax.vmap(project_i, (0,0), 0)(omega, alpha).sum(axis=0)

  fun = projection
  fun_i = jax.vmap(fun, (0), 0)(x_int)
  if num_diff == 0:
    return (fun_i)
  
  dfun_i = jax.jit(jax.vmap(jax.jacrev(fun), (0), 0))(x_int)
  if num_diff == 1:
    return (fun_i, dfun_i)
  
  ddfun_i = jax.jit(jax.vmap(jax.jacfwd(jax.jacrev(fun)), (0), 0))(x_int)
  if num_diff == 2:
    return (fun_i, dfun_i, ddfun_i)

### Remaining functions use discrete precomputed values and derivatives in order to make function that are more efficiently traceable for JAX. Can require a lot of memory, if the max-derivative-order is high.

@jax.custom_jvp
def _compiled_function_1(x, f, df):
  return f

@_compiled_function_1.defjvp
def _compiled_function_1_jvp(primals, tangents):
  x, f, df = primals
  x_dot, _, _ = tangents
  primal_out = _compiled_function_1(x, f, df)
  tangent_out = df @ x_dot
  return primal_out, tangent_out
compiled_function_1 = jax.jit(jax.vmap(_compiled_function_1, (None, 0, 0), 0))

@jax.custom_jvp
def _compiled_function_2(x, f, df, ddf):
  return f

@_compiled_function_2.defjvp
def _compiled_function_2_jvp(primals, tangents):
  x, f, df, ddf = primals
  x_dot, _, _, _ = tangents
  primal_out = _compiled_function_2(x, f, df, ddf)
  tangent_out = df @ x_dot + (ddf @ x) @ x_dot
  return primal_out, tangent_out
compiled_function_2 = jax.jit(jax.vmap(_compiled_function_2, (None, 0, 0, 0), 0))



# Copyright (C) 2024 Tobias Bode
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

import time
import sys

import jax
from jax import lax, config
import jax.numpy as jnp
import flax

from autopdex import seeder, geometry, solver, solution_structures, plotter, utility, models, spaces, assembler

config.update("jax_enable_x64", True)
config.update("jax_compilation_cache_dir", './cache')


### Definition of geometry and boundary conditions
# Cook's membrane, see e.g. https://link.springer.com/article/10.1007/s00466-017-1405-4

### Definition of geometry and boundary conditions
# Positive smooth distance functions that select the specific boundary segments
pts = [[0.,0.], [48.,44.], [48.,60.], [0.,44.]]
psdf_bottom = lambda x: geometry.psdf_trimmed_line(x, pts[0], pts[1])
psdf_right = lambda x: geometry.psdf_trimmed_line(x, pts[1], pts[2])
psdf_top = lambda x: geometry.psdf_trimmed_line(x, pts[2], pts[3])
psdf_left = lambda x: geometry.psdf_trimmed_line(x, pts[3], pts[0])
psdfs = [psdf_bottom, psdf_right, psdf_top, psdf_left]

# Positive smooth distance funcion of full domain
sdf_cooks = lambda x: geometry.sdf_convex_polygon_2d(x, pts)
psdf_cooks = lambda x: geometry.only_positive(sdf_cooks(x))
psdf = lambda x: jnp.asarray([psdf_cooks(x), psdf_cooks(x)])

# Boundary condition functions
homogeneous = lambda x: 0.
normals = lambda x: geometry.normals_from_normalized_sdf(x, psdf_cooks)

# Field selectors for boundary conditions
selection_dirichlet_1 = lambda x: jnp.asarray([1.,0.,0.,0.,0.,0.])
selection_dirichlet_2 = lambda x: jnp.asarray([0.,1.,0.,0.,0.,0.])
def selection_neumann_1(x):
  normal_x = normals(x)[0]
  normal_y = normals(x)[1]
  return jnp.asarray([0.,0.,normal_x,normal_y,0.,0.])
def selection_neumann_2(x):
  normal_x = normals(x)[0]
  normal_y = normals(x)[1]
  return jnp.asarray([0.,0.,0.,0.,normal_x,normal_y])

# Assignment of boundary conditions (same ordering as in psdfs)
load = lambda x: 4.0
bc_1 = [homogeneous, homogeneous, homogeneous, homogeneous]
bc_2 = [homogeneous, load, homogeneous, homogeneous]
a_1 = [selection_neumann_1, selection_neumann_1, selection_neumann_1, selection_dirichlet_1]
a_2 = [selection_neumann_2, selection_neumann_2, selection_neumann_2, selection_dirichlet_2]

### Transfinite interpolation of boundary segments
b_coeff = lambda x: jnp.asarray([geometry.transfinite_interpolation(x, psdfs, a_1),
                                   geometry.transfinite_interpolation(x, psdfs, a_2)])
b_cond = lambda x: jnp.asarray([geometry.transfinite_interpolation(x, psdfs, bc_1),
                                geometry.transfinite_interpolation(x, psdfs, bc_2)])


### Initialization and generation of nodes and integration points
# Regular node placement
# nod_spacing = 0.675
nod_spacing = 2.
(x_nodes, n_nodes, region_size) = seeder.regular_in_psdf(sdf_cooks, pts[0], pts[2], nod_spacing, atol=-1e-8)

# Gauss integration on background grid
int_spacing = nod_spacing
(x_int, w_int, n_int) = seeder.gauss_points_in_psdf(sdf_cooks, pts[0], pts[2], int_spacing, order=2)

# Support radius for moving least square ansatz
support_radius = 2.71 * nod_spacing

# # ### Visualizer for boundary functions
# fun = lambda x: b_cond(x)[1]
# plotter.pv_function_plot(x_int, fun)
# sys.exit()

# Definition of pde
lam = 100.
mu = 40.
Em = mu * (3 * lam + 2 * mu) / (lam + mu)
nu = lam / (2 * (lam + mu))
youngs_mod_fun = lambda x: Em
poisson_ratio_fun = lambda x: nu
# pde = models.linear_elasticity_fos(youngs_mod_fun = youngs_mod_fun, poisson_ratio_fun = poisson_ratio_fun, mode='plain strain', spacing=support_radius)
pde = models.hyperelastic_steady_state_fos(models.neo_hooke, youngs_mod_fun = youngs_mod_fun, poisson_ratio_fun = poisson_ratio_fun, mode='plain strain', spacing=support_radius)


### Neighborhood search
(num_neighbors, max_neighbors, min_neighbors, neighbor_list) = utility.search_neighborhood(x_nodes, x_int, support_radius)


### Setting
n_fields = 6
dofs0 = jnp.zeros((n_nodes, n_fields))
static_settings = flax.core.FrozenDict({
  'solution space': ('mls',),
  'shape function mode': 'compiled',
  'number of fields': (n_fields,),
  'assembling mode': ('sparse',),
  'maximal number of neighbors': (max_neighbors,),
  'order of basis functions': (2,),
  'weight function type': ('gaussian',),
  'variational scheme': ('least square pde loss',),
  'model': (pde,),
  'solution structure': ('first order set',),
  'boundary coefficients': (b_coeff,),
  'boundary conditions': (b_cond,),
  'psdf': (psdf,),
  'solver type': 'newton',
  'solver backend': 'pardiso',
  'solver': 'lu',
  'verbose': 2,
  'connectivity': (utility.jnp_to_tuple(neighbor_list),),
})
settings = {
  'node coordinates': x_nodes,
  'integration coordinates': (x_int,),
  'integration weights': (w_int,),
  'support radius': (support_radius,),
  'beta': (3,),
}


# Precompute shape functions
settings = solution_structures.precompile(dofs0, settings, static_settings)

### Load stepping in order to provide better initial guesses for the newton method
dofs = dofs0
multiplier = 0.0
n_step = 5
for i in range(n_step):
  # Update boundary conditions
  print('Start load step ', i+1)
  multiplier = multiplier + 1 / n_step
  load_mult = lambda x: multiplier * load(x)
  bc_2 = [homogeneous, load_mult, homogeneous, homogeneous]
  b_cond = lambda x: jnp.asarray([geometry.transfinite_interpolation(x, psdfs, bc_1),
                                  geometry.transfinite_interpolation(x, psdfs, bc_2)])
  settings['compiled bc'] = (solution_structures.precompute_coupled_boundary(b_cond, b_coeff, x_int, num_diff=1),)
  dofs, _ = solver.solver(dofs, settings, static_settings)



### Preperation of postprocessing data
local_dofs = dofs[neighbor_list]

def post_fun(x, itt, local_dofs, settings, static_settings):
  return solution_structures.solution_structure(x, itt, local_dofs, settings, static_settings, 0)

post_fun_vj = jax.jit(jax.vmap(post_fun, (0,0,0,None,None)), static_argnames=['static_settings'])
data = post_fun_vj(x_int, jnp.arange(n_int), local_dofs, settings, static_settings)

x_vis = x_int + jnp.transpose(jnp.asarray([data[:,0],data[:,1]]))

# Plotting with pyvista
plotter.pv_plot(x_vis, data[:,0], export_vtk=False, show=True, data_on_z=False)
plotter.pv_plot(x_vis, data[:,1], export_vtk=False, show=True, data_on_z=False)
plotter.pv_plot(x_vis, data[:,2], export_vtk=False, show=True, data_on_z=False)
plotter.pv_plot(x_vis, data[:,3], export_vtk=False, show=True, data_on_z=False)
plotter.pv_plot(x_vis, data[:,4], export_vtk=False, show=True, data_on_z=False)
plotter.pv_plot(x_vis, data[:,5], export_vtk=False, show=True, data_on_z=False)



# geometry.py
# Copyright (C) 2024 Tobias Bode
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

"""
This module contains geometry related functions that can be used for seeding nodes and integration points, impose boundary conditions and to select nodes and elements.
The main functionality relies on distance functions. Zur Konstruktion von benutzerspezifischen distance functions gibt es die Rvachev-Operationen,
es gibt aber auch einige vorgefertigte signed distance functions und positive signed distance functions. Die Distanzfunktionen eignen sich zum Beispiel, 
um zu überprüfen, ob Knoten innerhalb bestimmter Regionen oder auf spezifizierten Oberflächen liegen. Die glatten Abstandsfunktionen eigenen sich außerdem zur
Konstruktion von Lösungsstrukturen.
"""

import jax
import jax.numpy as jnp

from autopdex.utility import jit_with_docstring

# ToDo:
# - reader for stl data

### R function operations that preserve firt order normalization
def r_equivalence(o1, o2):
  """
  Computes the logical NOT XOR operation by combining two positively evaluated smooth distance functions.
  
  Parameters:
    o1 (float): The first operand for the logical "not xor" operation.
    o2 (float): The second operand for the logical "not xor" operation.

  Returns:
    float: The result of the logical "not xor" operation, calculated using 
        the formula o1 * o2 / (o1 + o2).
  """
  return o1 * o2 / (o1 + o2)

def r_conjunction(o1, o2):
  """
  Computes the logical AND operation using smooth distance functions.
  
  Parameters:
    o1 (float): The first operand for the logical "and" operation.
    o2 (float): The second operand for the logical "and" operation.

  Returns:
    float: The result of the logical "and" operation, calculated using 
        the formula o1 + o2 - sqrt(o1**2 + o2**2).
  """
  return o1 + o2 - jnp.sqrt(o1**2 + o2**2)

def r_disjunction(o1, o2):
  """
  Computes the logical OR operation using smooth distance functions.
  
  Parameters:
    o1 (float): The first operand for the logical "or" operation.
    o2 (float): The second operand for the logical "or" operation.

  Returns:
    float: The result of the logical "or" operation, calculated using 
        the formula o1 + o2 + sqrt(o1**2 + o2**2).
  """
  return o1 + o2 + jnp.sqrt(o1**2 + o2**2)

def r_trimming(o, t):
  """
  Trimming operations.

  Parameters:
    o (float): (Possibly signed) smooth distance function.
    t (float): Trimming function.

  Returns:
    float: Positive smooth distance function with zero set 
        as intersection of zero set of o and positive part of t.
  """
  return jnp.sqrt(o**2 + (jnp.sqrt(t**2 + o**4) - t)**2 / 4)

def signed_to_positive(o):
  """Absolute value using the formula jnp.sqrt(o**2)."""
  return jnp.sqrt(o**2)

def only_positive(o):
  """Set negative regions to zero by manipulating values using the formula (o + jnp.sqrt(o**2)) / 2."""
  return (o + jnp.sqrt(o**2)) / 2

def first_order_normalization(o, grad):
  """Normalizes an unnormalized distance function o with its gradient grad."""
  return o / jnp.sqrt(o**2 + jnp.dot(grad,grad))

def normals_from_normalized_sdf(x, sdf):
  """
  Returns a smooth vector field that equals the surface normals where the normalized smooth distance function sdf equals 0
  
  Parameters:
    x (jnp.ndarray): The position at which the vector field shall be evaluated
    sdf (callable): The normalized smooth distance function with argument x

  Returns:
    jnp.ndarray
  """
  return - jax.jacrev(sdf)(x)


### Some helper functions
def centroid(x_p1, x_p2):
  """
  Mean of two points in n-dimensions:
  
  result = (x_p1 + x_p2) / 2
  """
  return (x_p1 + x_p2) / 2

def normal(x_p1, x_p2):
  """
  Normalized vector showing from x_p1 to x_p2.
  
  result = (x_p2 - x_p1) / jnp.linalg.norm(x_p2 - x_p1)
  """
  return (x_p2 - x_p1) / jnp.linalg.norm(x_p2 - x_p1)

def project(x, normal):
  """
  Projects vector x in the direction of vector normal.
  
  result = jnp.dot(x,normal) * normal
  """
  return jnp.dot(x,normal) * normal

def project_on_line(x_p1, x_p2, x_p3):
  """
  Projects x_p1 on the line going through x_p2 and x_p3.
  
  result = x_p2 + project(x_p1 - x_p2, normal(x_p2, x_p3))
  """
  return x_p2 + project(x_p1 - x_p2, normal(x_p2, x_p3))


### First order normalized signed smooth distance function of primitives
def sdf_infinite_line(x, x_p1, x_p2):
  """
  Normalized signed smooth distance function of an infinite line going through x_p1 and x_p2.

  Parameters:
    x (jnp.ndarray): The point where the distance is evaluated.
    x_p1 (jnp.ndarray): First point on the infinite line.
    x_p2 (jnp.ndarray): Second point on the infinite line.

  Returns:
    float: The signed distance from the point x to the infinite line defined by x_p1 and x_p2.
  """
  x_p1 = jnp.asarray(x_p1)
  x_p2 = jnp.asarray(x_p2)
  line_length = jnp.linalg.norm(x_p2 - x_p1)
  return ((x[0]-x_p1[0])*(x_p2[1]-x_p1[1])-(x[1]-x_p1[1])*(x_p2[0]-x_p1[0])) / line_length

def sdf_nd_sphere(x, xc, r):
  """
  Normalized signed distance function to n-dimensional sphere, positive in the interior.

  Parameters:
    x (jnp.ndarray): The point where the distance is evaluated.
    xc (jnp.ndarray): Center of the sphere.
    r (float): Radius of the sphere.

  Returns:
    float: The normalized signed distance from the point x to the n-dimensional sphere defined by center xc and radius r.
  """
  xc = jnp.asarray(xc)
  dxc = x - xc
  return (r**2 - jnp.dot(dxc,dxc)) / (2*r)

def sdf_nd_planes(x, x_p1, x_p2): 
  """
  Normalized signed smooth distance function, positive between points (1D), lines (2D), and n-dimensional planes (nD).

  Parameters:
    x (jnp.ndarray): The point where the distance is evaluated.
    x_p1 (jnp.ndarray): Point on the first plane.
    x_p2 (jnp.ndarray): Point on the second plane.

  Returns:
    float: The signed distance from the point x to the planes defined by x_p1 and x_p2.
  """
  x_p1 = jnp.asarray(x_p1)
  x_p2 = jnp.asarray(x_p2)

  xc = centroid(x_p1, x_p2)
  len = jnp.linalg.norm(x_p2 - x_p1)
  dxc = x - xc

  n_plane = normal(x_p1,x_p2)
  dist_to_mid_plane = jnp.dot(dxc,n_plane)
  
  return (len**2 / 4 - dist_to_mid_plane**2) / len

def sdf_normal_to_direction(x, x_p1, normal):
  """
  Normalized signed smooth distance function to a plane defined by one point and a normal vector, positive in the direction of the normal vector.

  Parameters:
    x (jnp.ndarray): The point where the distance is evaluated.
    x_p1 (jnp.ndarray): A point on the plane.
    normal (jnp.ndarray): The normal vector of the plane.

  Returns:
    float: The distance from the point x to the plane defined by x_p1 and the normal vector.
  """
  x_p1 = jnp.asarray(x_p1)
  normal = jnp.asarray(normal)

  dist = x - x_p1
  return jnp.dot(dist, normal)

def sdf_infinite_cylinder(x, x_p1, normal, radius): 
  """
  Normalized signed smooth distance function, positive within an infinite cylinder in 3D.

  Parameters:
    x (jnp.ndarray): The point where the distance is evaluated.
    x_p1 (jnp.ndarray): A point on the center line of the cylinder.
    normal (jnp.ndarray): The direction vector of the cylinder.
    radius (float): The radius of the cylinder.

  Returns:
    float: The signed distance from the point x to the infinite cylinder.
  """
  x_p1 = jnp.asarray(x_p1)
  normal = jnp.asarray(normal)

  x_projected_on_centerline = project_on_line(x, x_p1, x_p1+normal)
  d = x - x_projected_on_centerline
  return (radius**2 - jnp.dot(d,d)) / (2*radius)

def sdf_cylinder(x, x_0, normal, radius, length):
  """
  Normalized signed smooth distance function, positive within a cylinder in 3D.

  Parameters:
    x (jnp.ndarray): The point where the distance is evaluated.
    x_0 (jnp.ndarray): The starting point on the center line of the cylinder.
    normal (jnp.ndarray): The direction vector of the cylinder.
    radius (float): The radius of the cylinder.
    length (float): The length of the cylinder.

  Returns:
    float: The signed distance from the point x to the cylinder.
  """
  x_0 = jnp.asarray(x_0)
  normal = jnp.asarray(normal)

  inf_cylinder = sdf_infinite_cylinder(x, x_0, normal, radius)
  region_between_end_pts = sdf_nd_planes(x, x_0, x_0 + length * normal)
  return r_conjunction(inf_cylinder,region_between_end_pts)

def sdf_cylinder_extruded(x, x_0, normal, radius, length, t0, t1):
  """
  Normalized signed smooth distance function of a cylinder that is extruded in time (fourth dimension).

  Parameters:
    x (jnp.ndarray): The point where the distance is evaluated.
    x_0 (jnp.ndarray): The starting point on the center line of the cylinder.
    normal (jnp.ndarray): The direction vector of the cylinder.
    radius (float): The radius of the cylinder.
    length (float): The length of the cylinder.
    t0 (float): The starting time of the extrusion.
    t1 (float): The ending time of the extrusion.

  Returns:
    float: The signed distance from the point x to the extruded cylinder.
  """
  x_0 = jnp.asarray(x_0)
  normal = jnp.asarray(normal)

  cylinder_in_3d = sdf_cylinder(x[:3], x_0, normal, radius, length)
  region_between_t0_t1 = sdf_nd_planes(x, jnp.asarray([0.,0.,0.,t0]), jnp.asarray([0.,0.,0.,t1]))
  return r_conjunction(cylinder_in_3d, region_between_t0_t1)

def sdf_triangle_2d(x, x_p1, x_p2, x_p3):
  """
  Normalized signed smooth distance function to a triangle (2D) or infinite triprism (3D), positive in the interior.

  Parameters:
    x (jnp.ndarray): The point where the distance is evaluated.
    x_p1 (jnp.ndarray): First vertex of the triangle.
    x_p2 (jnp.ndarray): Second vertex of the triangle.
    x_p3 (jnp.ndarray): Third vertex of the triangle.

  Returns:
    float: The signed distance from the point x to the triangle or triprism.
  """
  x_p1 = jnp.asarray(x_p1)
  x_p2 = jnp.asarray(x_p2)
  x_p3 = jnp.asarray(x_p3)  
  
  projected_1 = project_on_line(x_p1, x_p2, x_p3)
  projected_2 = project_on_line(x_p2, x_p3, x_p1)
  projected_3 = project_on_line(x_p3, x_p1, x_p2)
  
  o1 = sdf_nd_planes(x, x_p1, projected_1)
  o2 = sdf_nd_planes(x, x_p2, projected_2)
  o3 = sdf_nd_planes(x, x_p3, projected_3)

  return r_conjunction(r_conjunction(o1,o2),o3)

def sdf_convex_polygon_2d(x, x_p_list):
  """
  Normalized signed smooth distance function to a convex polygon (2D), positive in the interior.

  Parameters:
    x (jnp.ndarray): The point where the distance is evaluated.
    x_p_list (list of jnp.ndarray): List of vertices of the polygon in counter-clockwise order.

  Returns:
    float: The signed distance from the point x to the convex polygon.
  """
  n_segments = len(x_p_list)
  psdf_d = (lambda t: sdf_infinite_line(t, x_p_list[1], x_p_list[0]))(x)
  
  for i in range(n_segments - 1):
    x_0 = x_p_list[-i]
    x_1 = x_p_list[-(i+1)]
    psdf_i = (lambda t: sdf_infinite_line(t, x_0, x_1))(x)
    psdf_d = r_conjunction(psdf_d, psdf_i)
  return psdf_d

def sdf_infinite_triprism_3d(x, x_p1, x_p2, x_p3):
  """
  Normalized signed smooth distance function to a triangle (2D) or infinite triprism (3D), positive in the interior.

  Parameters:
    x (jnp.ndarray): The point where the distance is evaluated.
    x_p1 (jnp.ndarray): First vertex of the triangle.
    x_p2 (jnp.ndarray): Second vertex of the triangle.
    x_p3 (jnp.ndarray): Third vertex of the triangle.

  Returns:
    float: The signed distance from the point x to the triangle or triprism.
  """
  return sdf_triangle_2d(x, x_p1, x_p2, x_p3)

# def sdf_triangle_3d(x, x_p1, x_p2, x_p3):
#   x_p1 = jnp.asarray(x_p1)
#   x_p2 = jnp.asarray(x_p2)
#   x_p3 = jnp.asarray(x_p3) 
#   return #ToDo

# def sdf_triprism(x, x_p1, x_p2, x_p3, length):
#   x_p1 = jnp.asarray(x_p1)
#   x_p2 = jnp.asarray(x_p2)
#   x_p3 = jnp.asarray(x_p3) 
#   return #ToDo

# def sdf_tetrahedron_3d(x, x_p1, x_p2, x_p3, x_p4):
#   x_p1 = jnp.asarray(x_p1)
#   x_p2 = jnp.asarray(x_p2)
#   x_p3 = jnp.asarray(x_p3) 
#   x_p4 = jnp.asarray(x_p4) 
#   return #ToDo

# def sdf_tetrahedron_extruded(x, x_p1, x_p2, x_p3, x_p4, t0, t1):
#   x_p1 = jnp.asarray(x_p1)
#   x_p2 = jnp.asarray(x_p2)
#   x_p3 = jnp.asarray(x_p3) 
#   x_p4 = jnp.asarray(x_p4) 
#   return #ToDo

def sdf_parallelogram_2d(x, x_p1, x_p2, x_p3):
  """
  Normalized signed smooth distance function to a parallelogram, positive in the interior.

  Usage for rectangle with vertices [[0,0],[0,1],[1,1],[1,0]]:
  sdf_parallelogram(x, [0,0], [0,1], [1,0])

  Parameters:
    x (jnp.ndarray): The point where the distance is evaluated.
    x_p1 (jnp.ndarray): First vertex of the parallelogram.
    x_p2 (jnp.ndarray): Second vertex of the parallelogram.
    x_p3 (jnp.ndarray): Third vertex of the parallelogram.

  Returns:
    float: The signed distance from the point x to the parallelogram.
  """
  x_p1 = jnp.asarray(x_p1)
  x_p2 = jnp.asarray(x_p2)
  x_p3 = jnp.asarray(x_p3) 

  projected_2 = project_on_line(x_p2, x_p3, x_p1)
  projected_3 = project_on_line(x_p3, x_p1, x_p2)
  
  o2 = sdf_nd_planes(x, x_p2, projected_2)
  o3 = sdf_nd_planes(x, x_p3, projected_3)

  return r_conjunction(o2,o3)

# def sdf_infinite_parallelogram_prism(x, x_p1, x_p2, x_p3):
#   x_p1 = jnp.asarray(x_p1)
#   x_p2 = jnp.asarray(x_p2)
#   x_p3 = jnp.asarray(x_p3) 
#   return #ToDo

def sdf_cuboid(x, x_p1, x_p2):
  """
  Normalized signed smooth distance function to a cuboid, positive in the interior.

  Parameters:
    x (jnp.ndarray): The point where the distance is evaluated.
    x_p1 (jnp.ndarray): The point with minimum coordinates of the cuboid.
    x_p2 (jnp.ndarray): The point with maximum coordinates of the cuboid.

  Returns:
    float: The signed distance from the point x to the cuboid.
  """
  x_p1 = jnp.asarray(x_p1)
  x_p2 = jnp.asarray(x_p2)

  project_x = project_on_line(x_p2, x_p1, x_p1 + jnp.asarray([1.,0.,0.]))
  project_y = project_on_line(x_p2, x_p1, x_p1 + jnp.asarray([0.,1.,0.]))
  project_z = project_on_line(x_p2, x_p1, x_p1 + jnp.asarray([0.,0.,1.]))
  
  o1 = sdf_nd_planes(x, x_p1, project_x)
  o2 = sdf_nd_planes(x, x_p1, project_y)
  o3 = sdf_nd_planes(x, x_p1, project_z)

  return r_conjunction(r_conjunction(o1,o2),o3)

# def sdf_parallelepiped_extruded(x, x_p1, x_p2, x_p3, x_p4, t0, t1):
#   x_p1 = jnp.asarray(x_p1)
#   x_p2 = jnp.asarray(x_p2)
#   x_p3 = jnp.asarray(x_p3) 
#   x_p4 = jnp.asarray(x_p4) 
#   return #ToDo

# def sdf_3d_ball_extruded(x, xc, r):
#   xc = jnp.asarray(xc)
#   return #ToDo


### Smooth positive distance function of primitives, first order normalized
def psdf_infinite_line(x, x_p1, x_p2):
  """
  Positive normalized smooth distance function of an infinite line going through x_p1 and x_p2.

  Parameters:
    x (jnp.ndarray): The point where the distance is evaluated.
    x_p1 (jnp.ndarray): First point on the infinite line.
    x_p2 (jnp.ndarray): Second point on the infinite line.

  Returns:
    float: The positive distance from the point x to the infinite line defined by x_p1 and x_p2.
  """
  x_p1 = jnp.asarray(x_p1)
  x_p2 = jnp.asarray(x_p2)

  dx1 = x - x_p1
  normal_vector_in_line_direction = normal(x_p1, x_p2)
  normal_component = dx1 - project(dx1, normal_vector_in_line_direction)
  return jnp.linalg.norm(normal_component)

def psdf_trimmed_line(x, x_p1, x_p2):
  """
  Positive normalized smooth distance function of a line segment going from x_p1 to x_p2.

  Parameters:
    x (jnp.ndarray): The point where the distance is evaluated.
    x_p1 (jnp.ndarray): Starting point of the line segment.
    x_p2 (jnp.ndarray): Ending point of the line segment.

  Returns:
    float: The positive distance from the point x to the line segment defined by x_p1 and x_p2.
  """
  x_p1 = jnp.asarray(x_p1)
  x_p2 = jnp.asarray(x_p2)

  omega = sdf_infinite_line(x, x_p1, x_p2)

  xc = centroid(x_p1, x_p2)
  d = x_p2 - x_p1
  len = jnp.linalg.norm(d)
  trim = sdf_nd_sphere(x, xc, len/2)

  return r_trimming(omega, trim)

def psdf_nd_sphere(x, xc, r):
  """
  Interior positive normalized smooth distance function of an n-dimensional sphere.

  Parameters:
    x (jnp.ndarray): The point where the distance is evaluated.
    xc (jnp.ndarray): Center of the sphere.
    r (float): Radius of the sphere.

  Returns:
    float: The positive distance from the point x to the region outside of a n-dimensional sphere.
  """
  xc = jnp.asarray(xc)
  return only_positive(sdf_nd_sphere(x, xc, r))

def psdf_nd_sphere_cutout(x, xc, r):
  """
  Exterior positive normalized smooth distance function of an n-dimensional sphere.

  Parameters:
    x (jnp.ndarray): The point where the distance is evaluated.
    xc (jnp.ndarray): Center of the sphere.
    r (float): Radius of the sphere.

  Returns:
    float: The positive distance from the point x to the filled n-dimensional sphere.
  """
  xc = jnp.asarray(xc)
  return only_positive(-sdf_nd_sphere(x, xc, r))

def psdf_infinite_cylinder(x, x_p1, normal, radius): 
  """
  Positive normalized smooth distance function to an infinite cylinder in 3D.

  Parameters:
    x (jnp.ndarray): The point where the distance is evaluated.
    x_p1 (jnp.ndarray): A point on the center line of the cylinder.
    normal (jnp.ndarray): The direction vector of the cylinder.
    radius (float): The radius of the cylinder.

  Returns:
    float: The positive distance from the point x to the infinite cylinder. Zero outside of the cylinder.
  """
  x_p1 = jnp.asarray(x_p1)
  normal = jnp.asarray(normal)
  return only_positive(sdf_infinite_cylinder(x, x_p1, normal, radius))

def psdf_cylinder(x, x_0, normal, radius, length):
  """
  Positive normalized smooth distance function to a cylinder in 3D.

  Parameters:
    x (jnp.ndarray): The point where the distance is evaluated.
    x_0 (jnp.ndarray): The starting point on the center line of the cylinder.
    normal (jnp.ndarray): The direction vector of the cylinder.
    radius (float): The radius of the cylinder.
    length (float): The length of the cylinder.

  Returns:
    float: The positive distance from the point x to the cylinder. Zero outside of the cylinder.
  """
  x_0 = jnp.asarray(x_0)
  normal = jnp.asarray(normal)
  return only_positive(sdf_cylinder(x, x_0, normal, radius, length))

def psdf_cylinder_extruded(x, x_0, normal, radius, length, t0, t1):
  """
  Positive normalized smooth distance function of a cylinder that is extruded in time.

  Parameters:
    x (jnp.ndarray): The point where the distance is evaluated.
    x_0 (jnp.ndarray): The starting point on the center line of the cylinder.
    normal (jnp.ndarray): The direction vector of the cylinder.
    radius (float): The radius of the cylinder.
    length (float): The length of the cylinder.
    t0 (float): The starting time of the extrusion.
    t1 (float): The ending time of the extrusion.

  Returns:
    float: The positive distance from the point x to the extruded cylinder. Zero outside of the cylinder.
  """
  x_0 = jnp.asarray(x_0)
  normal = jnp.asarray(normal)
  return only_positive(sdf_cylinder_extruded(x, x_0, normal, radius, length, t0, t1))

def psdf_triangle_2d(x, x_p1, x_p2, x_p3):
  """
  Interior positive normalized smooth distance function of a triangle in 2D and a tri-prism in 3D.

  Parameters:
    x (jnp.ndarray): The point where the distance is evaluated.
    x_p1 (jnp.ndarray): First vertex of the triangle.
    x_p2 (jnp.ndarray): Second vertex of the triangle.
    x_p3 (jnp.ndarray): Third vertex of the triangle.

  Returns:
    float: The positive distance from the point x to the region outside of the triangle or tri-prism.
  """
  x_p1 = jnp.asarray(x_p1)
  x_p2 = jnp.asarray(x_p2)
  x_p3 = jnp.asarray(x_p3) 
  return only_positive(sdf_triangle_2d(x, x_p1, x_p2, x_p3))

def psdf_infinite_triprism_3d(x, x_p1, x_p2, x_p3):
  """
  Interior positive normalized smooth distance function of a triangle in 2D and a tri-prism in 3D.

  Parameters:
    x (jnp.ndarray): The point where the distance is evaluated.
    x_p1 (jnp.ndarray): First vertex of the triangle.
    x_p2 (jnp.ndarray): Second vertex of the triangle.
    x_p3 (jnp.ndarray): Third vertex of the triangle.

  Returns:
    float: The positive distance from the point x to the region outside of the triangle or tri-prism.
  """
  x_p1 = jnp.asarray(x_p1)
  x_p2 = jnp.asarray(x_p2)
  x_p3 = jnp.asarray(x_p3) 
  return psdf_triangle_2d(x, x_p1, x_p2, x_p3)

# def psdf_triprism(x, x_p1, x_p2, x_p3, length):
#   x_p1 = jnp.asarray(x_p1)
#   x_p2 = jnp.asarray(x_p2)
#   x_p3 = jnp.asarray(x_p3) 
#   return #ToDo

def psdf_parallelogram(x, x_p1, x_p2, x_p3):
  """
  Positive normalized smooth distance function to a parallelogram, positive in the interior.

  Usage for rectangle with vertices [[0,0],[0,1],[1,1],[1,0]]:
  psdf_parallelogram(x, [0,0], [0,1], [1,0])

  Parameters:
    x (jnp.ndarray): The point where the distance is evaluated.
    x_p1 (jnp.ndarray): First vertex of the parallelogram.
    x_p2 (jnp.ndarray): Second vertex of the parallelogram.
    x_p3 (jnp.ndarray): Third vertex of the parallelogram.

  Returns:
    float: The positive distance from the point x to the region outside of the parallelogram.
  """
  x_p1 = jnp.asarray(x_p1)
  x_p2 = jnp.asarray(x_p2)
  x_p3 = jnp.asarray(x_p3) 
  return only_positive(sdf_parallelogram_2d(x, x_p1, x_p2, x_p3))

def psdf_arc_in_2d(x, xc, x_p1, x_p2):
  """
  First order normalized positive smooth distance function of a circular arc in 2D.

  Parameters:
    xc (jnp.ndarray): Center of the arc.
    x_p1 (jnp.ndarray): Start point of the arc.
    x_p2 (jnp.ndarray): End point of the arc (counter-clockwise).

  Returns:
    float: The normalized smooth distance from the point x to the circular arc.
  """
  x_p1 = jnp.asarray(x_p1)
  x_p2 = jnp.asarray(x_p2)
  xc = jnp.asarray(xc)
  [x1, y1] = x_p1
  [x2, y2] = x_p2
  radius = jnp.linalg.norm(xc-x_p1)

  # Distance and trimming function
  o = sdf_nd_sphere(x, xc, radius)
  t = ((x[0]-x1)*(y2-y1) - (x[1]-y1)*(x2-x1)) / jnp.sqrt((x2-x1)**2 + (y2-y1)**2)

  return r_trimming(o, t)

def psdf_polygon(x, x_p_list):
  """
  Positive normalized smooth distance function for a polygon in 2D.

  Parameters:
    x (jnp.ndarray): The point where the distance is evaluated.
    x_p_list (list of jnp.ndarray): List of edge coordinates of the polygon.

  Returns:
    float: The positive distance from the point x to the polygon boundary. (Positive inside and outside)
  """
  x_p_list = jnp.asarray(x_p_list)
  n_segments = x_p_list.shape[0]
  psdf_d = (lambda t: psdf_trimmed_line(t, x_p_list[0], x_p_list[1]))(x)
  
  def body_fun(i, psdf_d):
    x_0 = x_p_list[-i]
    x_1 = x_p_list[-(i+1)]
    psdf_i = (lambda t: psdf_trimmed_line(t, x_0, x_1))(x)
    return r_equivalence(psdf_d, psdf_i)
  return jax.lax.fori_loop(0, n_segments - 1, body_fun, psdf_d)


### Transfinite interpolation
def transfinite_interpolation(x, boundary_psdfs, boundary_conditions):
  """
  Transfinite interpolation using positive smooth distance functions (PSDFs) and boundary conditions.

  Parameters:
    x (jnp.ndarray): The point where the interpolation is evaluated.
    boundary_psdfs (list of callable): List of PSDF functions, each defining a boundary.
    boundary_conditions (list of callable): List of boundary condition functions corresponding to the PSDF functions.

  Returns:
    jnp.ndarray: The interpolated value at point x based on the boundary PSDFs and boundary conditions.

  Notes:
    This function uses an unstable version of transfinite interpolation at the boundary.
    Usually one can simply delete all integration points (and in case of collocation also nodes) on the boundary.
    A more stable but inefficient version (commented out) can be used if stability issues arise.
  """
  psdf_evaluated = jnp.stack([psdf_i(x) for psdf_i in boundary_psdfs])
  bc_evaluated = jnp.stack([bc_i(x) for bc_i in boundary_conditions])

  # Instable version (at boundary)
  psdf_inv = 1. / psdf_evaluated
  psdf_inv_sum = psdf_inv.sum()
  bc_d = jnp.dot(psdf_inv, bc_evaluated) / psdf_inv_sum
  return bc_d

  # # Stable version
  # denom = 0
  # for k in range(len(psdf_evaluated)):
  #   wk = 1
  #   for (j,psdf_j) in enumerate(psdf_evaluated):
  #     if k != j:
  #       wk = wk * psdf_j
  #   denom = denom + wk

  # interpolated = 0
  # for i in range(len(psdf_evaluated)):
  #   wi = 1
  #   for (j,psdf_j) in enumerate(psdf_evaluated):
  #     if i != j:
  #       wi = wi * psdf_j
  #   interpolated = interpolated + wi * bc_evaluated[i]
  # return interpolated / denom

def psdf_unification(x, boundary_psdfs):
  """
  Unification of multiple positive smooth distance functions (PSDFs).

  Parameters:
    x (jnp.ndarray): The point where the unification is evaluated.
    boundary_psdfs (list of callable): List of PSDF functions to be unified.

  Returns:
    float: The unified PSDF value at point x.

  Notes:
    The unification process combines the PSDFs using the r_equivalence relation.
  """
  psdf_d = boundary_psdfs[0](x)
  for psdf_i in boundary_psdfs[1:]:
    psdf_d = r_equivalence(psdf_d, psdf_i(x))

  return psdf_d


### Mesh related functions
@jit_with_docstring()
def triangle_area(x_edges):
  """
  Compute the area of a triangle given its vertices (jitted).

  Parameters:
    x_edges (jnp.ndarray): Array of shape (3, 2) containing the coordinates of the triangle's vertices.

  Returns:
    float: The area of the triangle.
  """
  x_0 = x_edges[0]
  x_1 = x_edges[1]
  x_2 = x_edges[2]

  x_2_projected = project_on_line(x_2, x_0, x_1)
  l = jnp.linalg.norm(x_1-x_0)
  h = jnp.linalg.norm(x_2 - x_2_projected)
  area = h * l / 2
  return area

@jit_with_docstring()
def triangle_areas(x_nodes, elements):
  """
  Compute the areas of multiple triangles given their vertices and connectivity (jitted).

  Parameters:
    x_nodes (jnp.ndarray): Array of shape (n, 2) containing the coordinates of the nodes.
    elements (jnp.ndarray): Array of shape (m, 3) containing the indices of the nodes forming each triangle.

  Returns:
    jnp.ndarray: Array of shape (m,) containing the areas of the triangles.
  """
  areas = jax.vmap(triangle_area, (0), 0)(x_nodes[elements])
  return areas

def in_sdf(x, sdf_fun):
  """
  Check if a point lies on the surface defined by a positive or signed distance function.

  Parameters:
    x (jnp.ndarray): The point to be checked.
    sdf_fun (callable): The distance function.

  Returns:
    bool: True if the point lies on the surface, False otherwise.
  """
  x = jnp.asarray(x)
  sdf = sdf_fun(x)
  return jnp.where(jnp.isclose(sdf, 0), True, False)

def in_sdfs(x, sdf_fun):
  """
  Check if multiple points lie on the surface defined by a positive or signed distance function.

  Parameters:
    x (jnp.ndarray): Array of points to be checked.
    sdf_fun (callable): The signed distance function.

  Returns:
      jnp.ndarray: Array of booleans indicating whether each point lies on the surface.
  """
  x = jnp.asarray(x)
  return jax.vmap(in_sdf, (0, None), 0)(x, sdf_fun)

def select_in_sdfs(x, sdf_fun):
  """
  Select the indices of points that lie on the surface defined by a positive or signed distance function.

  Parameters:
    x (jnp.ndarray): Array of points to be checked.
    sdf_fun (callable): The signed distance function.

  Returns:
    jnp.ndarray: Array of indices of points that lie on the surface.
  """
  x = jnp.asarray(x)
  return jnp.compress(in_sdfs(x, sdf_fun), jnp.arange(x.shape[0], dtype=jnp.int_), axis=0)

@jit_with_docstring()
def in_plane(x, x_p1, normal):
  """
  Check if a point lies in a plane defined by a point and a normal vector (jitted).

  Parameters:
    x (jnp.ndarray): The point to be checked.
    x_p1 (jnp.ndarray): A point on the plane.
    normal (jnp.ndarray): The normal vector of the plane.

  Returns:
    bool: True if the point lies in the plane, False otherwise.
  """
  sdf_fun = lambda x: sdf_normal_to_direction(x, x_p1, normal)
  return in_sdf(jnp.asarray(x), sdf_fun)

@jit_with_docstring()
def in_planes(x_q, x_p1, normal):
  """
  Check if multiple points lie in a plane defined by a point and a normal vector (jitted).

  Parameters:
    x_q (jnp.ndarray): Array of points to be checked.
    x_p1 (jnp.ndarray): A point on the plane.
    normal (jnp.ndarray): The normal vector of the plane.

  Returns:
    jnp.ndarray: Array of booleans indicating whether each point lies in the plane.
  """
  return jax.vmap(in_plane, (0, None, None), 0)(x_q, x_p1, normal)

def select_in_plane(x_q, x_p1, normal):
  """
  Select the indices of points that lie in a plane (non-jittable).

  Parameters:
    x_q (jnp.ndarray): Array of points to be checked.
    x_p1 (jnp.ndarray): A point on the plane.
    normal (jnp.ndarray): The normal vector of the plane.

  Returns:
    jnp.ndarray: Array of indices of points that lie in the plane.
  """
  return jnp.compress(in_planes(x_q, x_p1, normal), jnp.arange(x_q.shape[0], dtype=jnp.int_), axis=0)

@jit_with_docstring()
def on_point(x_q, x_p1):
  """
  Check if points are close to a given point (jitted).

  Parameters:
    x_q (jnp.ndarray): Array of points to be checked.
    x_p1 (jnp.ndarray): The reference point.

  Returns:
    jnp.ndarray: Array of booleans indicating whether each point is close to the reference point.
  """
  x_p1 = jnp.asarray(x_p1)
  x_q = jnp.asarray(x_q)
  
  def one_check(x_q, x_p1):
    sdf = jnp.linalg.norm(x_q - x_p1)
    return jnp.where(jnp.isclose(sdf, 0), True, False)
  
  map_check = lambda x_q, x_p1: jax.jit(jax.vmap(one_check, (0, None), 0))(x_q, x_p1)
  return map_check(x_q, x_p1)

def select_point(x_q, x_p1):
  """
  Select the index of a point that is close to a given point (non-jittable).

  Parameters:
    x_q (jnp.ndarray): Array of points to be checked.
    x_p1 (jnp.ndarray): The reference point.

  Returns:
    int: The index of the point that is close to the reference point.
  """
  return jnp.compress(on_point(x_q, x_p1), jnp.arange(x_q.shape[0], dtype=jnp.int_), axis=0)[0]

@jit_with_docstring()
def on_line(x, x_p1, x_p2):
  """
  Check if a point lies on a line segment defined by two points (jitted).

  Parameters:
    x (jnp.ndarray): The point to be checked.
    x_p1 (jnp.ndarray): The starting point of the line segment.
    x_p2 (jnp.ndarray): The ending point of the line segment.

  Returns:
    bool: True if the point lies on the line segment, False otherwise.
  """
  sdf_fun = lambda x: psdf_trimmed_line(x, x_p1, x_p2)
  return in_sdf(jnp.asarray(x), sdf_fun)

@jit_with_docstring()
def on_lines(x_q, x_p1, x_p2):
  """
  Check if multiple points lie on a line segment defined by two points (jitted).

  Parameters:
    x_q (jnp.ndarray): Array of points to be checked.
    x_p1 (jnp.ndarray): The starting point of the line segment.
    x_p2 (jnp.ndarray): The ending point of the line segment.

  Returns:
    jnp.ndarray: Array of booleans indicating whether each point lies on the line segment.
  """
  return jax.vmap(on_line, (0, None, None), 0)(x_q, x_p1, x_p2)

def select_on_line(x_q, x_p1, x_p2):
  """
  Select the indices of points that lie on a line segment defined by two points (non-jittable).

  Parameters:
    x_q (jnp.ndarray): Array of points to be checked.
    x_p1 (jnp.ndarray): The starting point of the line segment.
    x_p2 (jnp.ndarray): The ending point of the line segment.

  Returns:
    jnp.ndarray: Array of indices of points that lie on the line segment.
  """
  return jnp.compress(on_lines(x_q, x_p1, x_p2), jnp.arange(x_q.shape[0], dtype=jnp.int_), axis=0)

def elem_in_plane(x_nodes, x_p1, normal):
  """
  Check if an element lies in a plane defined by a point and a normal vector.

  Parameters:
    x_nodes (jnp.ndarray): Array of nodes of the element.
    x_p1 (jnp.ndarray): A point on the plane.
    normal (jnp.ndarray): The normal vector of the plane.

  Returns:
    bool: True if the element lies in the plane, False otherwise.
  """
  x_nodes = jnp.asarray(x_nodes)
  x_p1 = jnp.asarray(x_p1)
  normal = jnp.asarray(normal)
  return in_planes(x_nodes, x_p1, normal).all()

def select_elements_in_plane(x_nodes, surface_elements, x_p1, normal):
  """
  Select the indices of elements that lie in a plane.

  Parameters:
    x_nodes (jnp.ndarray): Array of nodes of the elements.
    surface_elements (jnp.ndarray): Array of surface elements.
    x_p1 (jnp.ndarray): A point on the plane.
    normal (jnp.ndarray): The normal vector of the plane.

  Returns:
    jnp.ndarray: Array of indices of elements that lie in the plane.
  """
  x_nodes = jnp.asarray(x_nodes)
  surface_elements = jnp.asarray(surface_elements)
  normal = jnp.asarray(normal)
  x_p1 = jnp.asarray(x_p1)
  return jax.vmap(elem_in_plane, (0, None, None), 0)(x_nodes[surface_elements], x_p1, normal)

def elem_on_line(x_nodes, x_p1, x_p2):
  """
  Check if an element lies on a line segment defined by two points.

  Parameters:
    x_nodes (jnp.ndarray): Array of nodes of the element.
    x_p1 (jnp.ndarray): The starting point of the line segment.
    x_p2 (jnp.ndarray): The ending point of the line segment.

  Returns:
    bool: True if the element lies on the line segment, False otherwise.
  """
  x_nodes = jnp.asarray(x_nodes)
  x_p1 = jnp.asarray(x_p1)
  x_p2 = jnp.asarray(x_p2)
  return on_lines(x_nodes, x_p1, x_p2).all()

def select_elements_on_line(x_nodes, surface_elements, x_p1, x_p2):
  """
  Select the indices of elements that lie on a line segment defined by two points.

  Parameters:
    x_nodes (jnp.ndarray): Array of nodes of the elements.
    surface_elements (jnp.ndarray): Array of surface elements.
    x_p1 (jnp.ndarray): The starting point of the line segment.
    x_p2 (jnp.ndarray): The ending point of the line segment.

  Returns:
    jnp.ndarray: Array of indices of elements that lie on the line segment.
  """
  x_nodes = jnp.asarray(x_nodes)
  surface_elements = jnp.asarray(surface_elements)
  x_p1 = jnp.asarray(x_p1)
  x_p2 = jnp.asarray(x_p2)
  return jax.vmap(elem_on_line, (0, None, None), 0)(x_nodes[surface_elements], x_p1, x_p2)

@jit_with_docstring()
def subelem_in_elem(subelem_to_find, in_elem):
  """
  Check if a sub-element (e.g. a line element) is contained within an element (e.g. triangular element) (jitted).

  Parameters:
    subelem_to_find (jnp.ndarray): The sub-element to be checked.
    in_elem (jnp.ndarray): The element to be checked against.

  Returns:
    bool: True if the sub-element is contained within the element, False otherwise.
  """
  return jnp.all(jnp.isin(subelem_to_find, in_elem))

def subelem_in_elems(subelem_to_find, in_elems):
  """
  Find the element that contains a given sub-element. Returns only one element (Only to be used for these cases).

  Parameters:
    subelem_to_find (jnp.ndarray): The sub-element to be found.
    in_elems (jnp.ndarray): Array of elements to be checked.

  Returns:
    jnp.ndarray: The element that contains the sub-element.
  """
  selection = jax.jit(jax.vmap(subelem_in_elem, (None, 0), 0))(subelem_to_find, in_elems)
  compression = jnp.compress(selection, in_elems, axis=0)[0]
  return compression

def subelems_in_elems(subelems_to_find, in_elems):
  """
  Find the elements that contain given sub-elements (one elem per sub_elem). 
  
  Can be used for detecting the domain elements that belong the given surface elements.

  Parameters:
    subelems_to_find (jnp.ndarray): Array of sub-elements to be found.
    in_elems (jnp.ndarray): Array of elements to be checked.

  Returns:
    jnp.ndarray: Array of elements that contain the sub-elements.
  """
  return jnp.asarray([subelem_in_elems(subelem, in_elems) for subelem in subelems_to_find])
  # return jax.vmap(subelem_in_elems, (0, None), 0)(subelems_to_find, in_elems)



# Copyright (C) 2024 Tobias Bode
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

def test_nonhomogeneous_dirichlet():

    import jax
    from jax import config
    import jax.numpy as jnp
    import numpy as np
    import flax
    import pygmsh

    from autopdex import seeder, geometry, solver, utility, models

    config.update("jax_enable_x64", True)


    ### Cook's membrane problem with non-homogeneous Dirichlet conditions nodally imposed

    ### Definition of geometry and boundary conditions
    pts = [[0.,0.], [48.,44.], [48.,60.], [0.,44.]]
    with pygmsh.geo.Geometry() as geom:
        geom.add_polygon(pts,mesh_size=50.0)
        mesh = geom.generate_mesh(order=1)

    # Import region mesh
    n_dim = 2
    x_nodes = jnp.asarray(mesh.points[:,:n_dim])
    n_nodes = x_nodes.shape[0]
    elements = jnp.asarray(mesh.cells_dict['triangle'])
    surface_elements = jnp.asarray(mesh.cells_dict['line'])

    # Selection of nodes for boundary condition
    dirichlet_left = geometry.in_planes(x_nodes, pts[0], [1.,0.])
    dirichlet_dofs_left = utility.dof_select(dirichlet_left, jnp.asarray([True, True]))
    dirichlet_conditions = jnp.zeros_like(dirichlet_dofs_left, dtype=jnp.float_)

    # Assignment of boundary conditions at right edge
    v_bar = lambda x: 0.1
    dirichlet_right = geometry.in_planes(x_nodes, pts[1], [1.,0.])
    dirichlet_dofs_right = utility.dof_select(dirichlet_right, jnp.asarray([False, True])) # Only impose y-direction
    x_right = x_nodes[dirichlet_right]
    v_right = jax.vmap(v_bar)(x_right)
    dirichlet_conditions = utility.mask_set(dirichlet_conditions, dirichlet_dofs_right, v_right)

    # All dirichlet dofs
    dirichlet_dofs = dirichlet_dofs_left + dirichlet_dofs_right

    # Generate domain integration points in mesh
    (x_int, w_int, n_int, domain_connectivity) = seeder.int_pts_in_tri_mesh(x_nodes, elements, order=1)
    max_neighbors_1 = domain_connectivity.shape[-1]

    # Definition of pde
    lam = 100.
    mu = 40.
    Em = mu * (3 * lam + 2 * mu) / (lam + mu)
    nu = lam / (2 * (lam + mu))
    youngs_mod_fun = lambda x: Em
    poisson_ratio_fun = lambda x: nu
    # pde_1 = models.linear_elasticity_weak(youngs_mod_fun, poisson_ratio_fun, 'plain strain')
    pde_1 = models.hyperelastic_steady_state_weak(models.neo_hooke, youngs_mod_fun, poisson_ratio_fun, 'plain strain')


    ### Setting
    n_fields = 2
    dofs0 = jnp.zeros((n_nodes, n_fields))
    static_settings = flax.core.FrozenDict({
    'solution space': ('fem simplex',),
    'shape function mode': 'direct',
    'number of fields': (n_fields,),
    'assembling mode': ('sparse',),
    'maximal number of neighbors': (max_neighbors_1,),
    'variational scheme': ('weak form galerkin',),
    'model': (pde_1,),
    'solution structure': ('nodal imposition',),
    'solver type': 'newton',
    'solver backend': 'scipy',
    'solver': 'lapack',
    'verbose': 0,
    'dirichlet dofs': utility.jnp_to_tuple(dirichlet_dofs),
    'connectivity': (utility.jnp_to_tuple(domain_connectivity),),
    })

    settings = {
    'node coordinates': x_nodes,
    'dirichlet conditions': dirichlet_conditions,
    'integration coordinates': (x_int,),
    'integration weights': (w_int,),
    }

    # Call solver
    dofs, _ = solver.solver(dofs0, settings, static_settings, tol=1e-8)

    # print(repr(dofs))
    assert jnp.allclose(dofs, jnp.asarray([[ 0.        ,  0.        ],
       [-0.01806664,  0.1       ],
       [-0.05176078,  0.1       ],
       [ 0.        ,  0.        ],
       [ 0.00834901,  0.02592409],
       [-0.02664544,  0.03728282]]))



# Copyright (C) 2024 Tobias Bode
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

import sys

import jax
from jax import config
import jax.numpy as jnp
import flax
import pyvista as pv
import numpy as np

from autopdex import seeder, geometry, solver, solution_structures, plotter, utility, models, spaces, assembler

config.update("jax_enable_x64", True)
config.update("jax_compilation_cache_dir", './cache')


### Model problem from https://docs.nvidia.com/deeplearning/modulus/modulus-sym/user_guide/theory/advanced_schemes.html#exact-boundary-condition-imposition
# MLS solution structure, comparison of galerkin and least square variational methods


### Definition of geometry and boundary conditions
# Positive smooth distance functions that select the specific boundary segments
pts = [[0.,0.], [0.1,0.], [0.1,0.1], [0.,0.1]]
psdf_bottom = lambda x: geometry.psdf_trimmed_line(x, pts[0], pts[1])
psdf_right = lambda x: geometry.psdf_trimmed_line(x, pts[1], pts[2])
psdf_top = lambda x: geometry.psdf_trimmed_line(x, pts[2], pts[3])
psdf_left = lambda x: geometry.psdf_trimmed_line(x, pts[3], pts[0])
psdf_rect = lambda x: geometry.psdf_parallelogram(x, pts[0], pts[3], pts[1])
psdfs = [psdf_bottom, psdf_right, psdf_top, psdf_left]

# Positive smooth distance funcion of full domain
psdf = lambda x: jnp.asarray([psdf_rect(x), psdf_rect(x)])

# Field selectors for boundary conditions
selection_dirichlet_1 = lambda x: jnp.asarray([0.,1.,0.])
selection_dirichlet_2 = lambda x: jnp.asarray([0.,0.,1.])

# Assignment of boundary conditions (same ordering as in psdfs)
homogeneous = lambda x: 0.
one = lambda x: 1.
bc_1 = [homogeneous, homogeneous, one, homogeneous]

### Transfinite interpolation of boundary segments
b_coeff = lambda x: jnp.asarray([selection_dirichlet_1(x),
                                 selection_dirichlet_2(x)])
b_cond = lambda x: jnp.asarray([geometry.transfinite_interpolation(x, psdfs, bc_1),
                                homogeneous(x)])


### Initialization and generation of nodes and integration points
# Regular node placement
nod_spacing = 0.002
(x_nodes, n_nodes, region_size) = seeder.regular_in_psdf(psdf_rect, pts[0], pts[2], nod_spacing)

# Gauss integration on background grid
int_spacing = nod_spacing
(x_int, w_int, n_int) = seeder.gauss_points_in_psdf(psdf_rect, pts[0], pts[2], int_spacing, order=2)

# Support radius for moving least square ansatz
support_radius = 3.01 * nod_spacing

# # ### Visualizer for boundary functions
# fun = lambda x: b_cond(x)[1]
# plotter.pv_function_plot(x_int, fun)
# sys.exit()

# Definition of pde
pde = models.navier_stokes_incompressible_steady(dynamic_viscosity=0.01)


### Neighborhood search
(num_neighbors, max_neighbors, min_neighbors, neighbor_list) = utility.search_neighborhood(x_nodes, x_int, support_radius)


### Setting
n_fields = 3
dofs0 = jnp.zeros((n_nodes, n_fields))
static_settings = flax.core.FrozenDict({
  'solution space': ('mls',),
  'shape function mode': 'compiled',
  'number of fields': (n_fields,),
  'assembling mode': ('sparse',),
  'maximal number of neighbors': (max_neighbors,),
  'order of basis functions': (2,),
  'weight function type': ('gaussian',),

  # Note the inaccuracy in the pressure field with least square variational scheme
  # 'variational scheme': ('least square pde loss',),
  'variational scheme': ('strong form galerkin',),

  'model': (pde,),
  'solution structure': ('second order set',),
  'boundary coefficients': (b_coeff,),
  'boundary conditions': (b_cond,),
  'psdf': (psdf,),
  'solver type': 'newton',
  'solver backend': 'pardiso',
  'solver': 'lu',
  'verbose': 1,
  'connectivity': (utility.jnp_to_tuple(neighbor_list),),
})
settings = {
  'beta': (3,),
  'node coordinates': x_nodes,
  'integration coordinates': (x_int,),
  'integration weights': (w_int,),
  'support radius': (support_radius,),
}


### Optional: Precompute shape functions and pass them to e.g. static_settings...
settings = solution_structures.precompile(dofs0, settings, static_settings)


### Call solver
dofs, _ = solver.solver(dofs0, settings, static_settings, newton_tol=1e-7) #, tol=1e-9, maxiter=10000


### Preperation of postprocessing data
local_dofs = dofs[neighbor_list]
post_fun_vj = jax.jit(jax.vmap(solution_structures.solution_structure, (0,0,0,None,None,None)), static_argnames=['static_settings', 'set'])
data = post_fun_vj(x_int, jnp.arange(n_int), local_dofs, settings, static_settings, 0)

# Interactive plotting with pyvista
plotter.pv_plot(x_int, data[:,0], export_vtk=False, show=True, data_on_z=False)
plotter.pv_plot(x_int, data[:,1], export_vtk=False, show=True, data_on_z=False)
plotter.pv_plot(x_int, data[:,2], export_vtk=False, show=True, data_on_z=False)



# utility.py
# Copyright (C) 2024 Tobias Bode
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

"""
This module contains some useful functions, including:

- Wrapper to nearest neighbor algorithm (e.g. for moving least squares)
- Degree of freedom (DOF) selection for boundary conditions
- Compute condition number and check symmetry of tangent matrix
- Functions for manipulating arrays
"""

from functools import partial, wraps

import jax
import jax.numpy as jnp
import numpy as np
from jax.experimental import sparse


def jit_with_docstring(static_argnames=None):
    def decorator(fun):
        # Define the jitted function with optional static_argnames
        if static_argnames is not None:
            jitted_fun = partial(jax.jit, static_argnames=static_argnames)(fun)
        else:
            jitted_fun = jax.jit(fun)

        # Create the wrapper function
        @wraps(fun)
        def wrapper(*args, **kwargs):
            return jitted_fun(*args, **kwargs)

        return wrapper
    return decorator

def search_neighborhood(x_nodes, x_query, support_radius):
  """
  Neighbor search within a radius based on scipy's KDTree.

  Args:
    x_nodes (array): Coordinates of the nodes.
    x_query (array): Coordinates of the query points.
    support_radius (float): Radius within which to search for neighbors.

  Returns:
    tuple: 
      - num_neighbors (array): Number of neighbors for each query point.
      - max_neighbors (int): Maximum number of neighbors.
      - min_neighbors (int): Minimum number of neighbors.
      - neighbor_list (array): List of neighbors for each query point.
    """
  import scipy

  tree = scipy.spatial.cKDTree(x_nodes)
  num_neighbors = jnp.asarray(tree.query_ball_point(x_query, support_radius, return_sorted=False, return_length=True))
  max_neighbors = num_neighbors.max().item()
  min_neighbors = num_neighbors.min().item()
  _, ii = tree.query(x_query, k=max_neighbors)
  neighbor_list = jnp.asarray(ii)

  print('Neighbor search finished. Min/max neighbors: ', (min_neighbors, max_neighbors))
  return (num_neighbors, max_neighbors, min_neighbors, neighbor_list)

@jit_with_docstring(static_argnames=['static_settings'])
def get_condition_number(dofs, settings, static_settings):
  """
  Computes the condition number of the assembled tangent matrix.

  Notes:
   - Computes a dense tangent. Therefore only useful for small problems.
  
  Args:
    dofs (jnp.ndarray): Degrees of freedom.
    settings (dict): Settings for the assembly process.
    static_settings (dict): Static settings.

  Returns:
    float: Condition number of the tangent matrix.
  """
  from autopdex import assembler
  (eig, _) = jnp.linalg.eig(assembler.assemble_tangent(dofs, settings, static_settings).todense())
  conditioning = eig[0] / eig[-1]
  return conditioning

@jit_with_docstring(static_argnames=['static_settings'])
def symmetry_check(dofs, settings, static_settings):
  """
  Checks the symmetry of the assembled tangent matrix.

  Args:
    dofs (jnp.ndarray): Degrees of freedom.
    settings (dict): Settings for the assembly process.
    static_settings (dict): Static settings.

  Returns:
    float: Sum of absolute differences between the matrix and its transpose.
  """
  from autopdex import assembler

  mat = assembler.assemble_tangent(dofs, settings, static_settings)
  return sparse.sparsify(jnp.sum)(sparse.sparsify(jnp.abs)(mat - sparse.bcoo_transpose(mat, permutation=(1,0))))

@jit_with_docstring()
def dof_select(dirichlet_nodes, selected_fields):
  """
  DOF selection for the nodal imposition of boundary conditions (jitted).
  
  Input:
    dirichlet_nodes: list of bools, indicating, which nods are dirichlet_nods
    selected_fields: as list of bools

  Returns:
    dirichlet_dofs: list of bools, indicating, which dofs are dirichlet_dofs
  """
  return jnp.outer(dirichlet_nodes, selected_fields)

def jnp_to_tuple(jnp_array):
  """
  Converts a JAX array to a tuple.

  Args:
    jnp.ndarray: JAX array to convert.

  Returns:
    tuple: Converted tuple.
  """
  as_list = np.asarray(jnp_array).tolist()
  try:
    return tuple(jnp_to_tuple(i) for i in as_list)
  except TypeError:
    return as_list

def mask_set(array, selection, values):
  """
  Changes values in a JAX array based on a boolean mask.

  Args:
    jnp.ndarray: JAX array to change.
    jnp.ndarray: Boolean mask.
    jnp.ndarray: Values to insert where the mask is true.

  Returns:
    array: Modified JAX array.
  """
  flat_array = array.flatten()
  idx = jnp.arange(flat_array.shape[0])
  flat_array = flat_array.at[idx[jnp.asarray(selection.flatten())]].set(values)
  return flat_array.reshape(array.shape)

def to_jax_function(subexpr, reduced_expr):
  """
  Generation of a pure JAX function based on subexpressions and reduced return-value. Can be combined with sympys common subexpression eliminator. 

  Return value of the function will be an jnp.ndarray  
  Warning: Works currently only functions with basic operations. Special functions as exp() may not be jax-conforming...
  Warning: arguments and function name have to be set manually. The function may have to be adjusted manually.
  """
  assignment_statements = []
  for var, expr in subexpr:
      assignment_statements.append(f"{var} = {expr}")

  # Create the function body
  function_body = "\n  ".join(assignment_statements)

  # Combine everything into a function definition
  function_code = "def shape_functions(xi):\n"
  function_code += f"  {function_body}\n"
  function_code += f"  return jnp.asarray({reduced_expr})\n"

  # Write the function code to a Python file
  with open('lagrange_polynomials.py', 'w') as f:
      f.write(function_code)

  # Print the generated function definition
  print(function_code)

@jit_with_docstring(static_argnames=['fun', 'derivative_order', 'argnum'])
def jacfwd_upto_n_scalar_args(fun, args, derivative_order, argnum):
    """
    Computes all up to the nth derivative of a function `f` using a single pass.

    The function is designed for the use in sensitivity analysis, where the primal evaluation can be very expensive.
    By using recursive jvps, the function computes all the derivatives in pass and does not rely on XLA to delete duplicated code.
    
    This function supports computing derivatives of any order with respect to individual or multiple 
    scalar arguments in a tuple. The derivatives are returned either as a single value or as a collection 
    of values, depending on the input type of `argnum`.

    Args:
        fun (Callable): The function for which derivatives are to be computed. This function should accept 
            a sequence of scalar inputs and return a scalar output or jnp.ndarray.
        args (tuple): A tuple of input values at which the derivatives are evaluated. The length of `args` 
            should match the number of arguments expected by `fun`.
        derivative_order (int): The order of the derivatives to compute. For instance, `n=2` computes the value, first and second derivative.
        argnum (Union[int, tuple]): The index or indices of the argument(s) with respect to which the derivative 
            is computed. If an integer is provided, the derivative is computed with respect to a single argument.
            If a tuple of integers is provided, the derivative is computed with respect to multiple arguments.

    Returns:
        jnp.ndarray:
        - The jnp.ndarray has the following order: results[i][j][k]
        - i (optional) selects the argument in the argnum tuple. This dimension dissapears in case argnum is a scalar
        - j is the j-th derivative order
        - k an further indices come from the output shape of the example function

    Example:
      def example_function(x, y):
        return jnp.asarray([x**3, 2*x*y**3, x*y])

      args = (2.0, 3.0)  # Example input values for x and y
      order = 4  # Up to the fourth derivative
      argnum = (0, 1) # Differentiate with respect to both x and y
      result = jacfwd_upto_n_scalar_args(example_function, args, order, argnum)
      print("Derivatives:", result)
      print("Derivatives:", result.shape) # Output: (2, 5, 3), like the dimensions of (argnum, 1+order, output_of_example_fun)
    """
    def nth_derivative_as_tuple(f, x, n, index):
        # Set the direction for differentiation for the specified argumentz
        tangent_directions = jnp.asarray([jnp.zeros_like(xi) for xi in x])
        tangent_directions = tangent_directions.at[index].set(1)

        if n == 0:
            return f(*x)
        elif n == 1:
            return jax.jvp(f, x, tuple(tangent_directions))
        else: # Recursive call of jvp
            df = lambda *y: nth_derivative_as_tuple(f, y, n-1, index)
            values = jax.jvp(df, x, tuple(tangent_directions))
            return values[0] + (values[-1][-1],)

    if isinstance(argnum, int):
        return jnp.asarray(nth_derivative_as_tuple(fun, args, derivative_order, argnum))
    elif isinstance(argnum, tuple):
        # Use jax.vmap to compute derivatives for each index in argnum
        fun_vmap = lambda idx: nth_derivative_as_tuple(fun, args, derivative_order, idx)
        return jnp.asarray(jax.vmap(fun_vmap)(jnp.asarray(argnum)))
    else:
        raise TypeError("argnum must be an integer or a tuple of integers.")

@jit_with_docstring(static_argnames=['fun', 'n'])
def jacfwd_upto_n_one_vector_arg(fun, x, n):
    """
    Computes up to the nth derivative of a function `fun` with respect to a vector argument `x` using a single pass.

    This function is tailored for sensitivity analysis, where the primal evaluation can be very expensive. By using
    recursive Jacobian-vector products (JVPs), this function computes all derivatives in one pass, avoiding redundant
    computations.

    The function supports computing derivatives of any order with respect to a single `jnp.ndarray` argument `x`. The derivatives
    are returned as a tuple, where each entry corresponds to the function's value or a derivative of increasing order.

    Args:
        fun (Callable): The function for which derivatives are to be computed. This function should accept a single `jnp.ndarray` 
            argument `x` and return a scalar output or `jnp.ndarray`.
        x (jnp.ndarray): The vector input at which the derivatives are evaluated.
        n (int): The order of the derivatives to compute. For instance, `n=3` computes the value, first, second, and third derivatives.

    Returns:
        tuple: A tuple containing the function value and its derivatives up to the nth order. 
        The structure of the tuple is as follows:
        - The first entry is the value of the function `fun(x)`.
        - Subsequent entries are the first, second, and higher-order derivatives.
        - For a function returning a multi-dimensional array, the array dimensions are appended to derivative dimensions (see example below)

    Example:
      def example_fun(x): # input shape (2,)
          return jnp.asarray([
              [x[0], x[1], x[0] + x[1], x[0] * x[1]],
              [x[1] - x[0], x[0] / (x[1] + 1e-5), x[0]**2, x[1]**2],
              [x[0] + 1, x[1] + 1, x[0] * x[0], x[1] * x[1]] 
          ]) # output shape (3, 4)

      x = jnp.asarray([1.0, 2.0])  # Example vector input
      n = 3  # Up to the third derivative
      derivatives = jacfwd_upto_n_one_vector_arg(example_fun, x, n)

      print("Function value:", derivatives[0].shape) # shape (3, 4)
      print("First derivative:", derivatives[1].shape)  # shape (2, 3, 4)
      print("Second derivative:", derivatives[2].shape) # shape (2, 2, 3, 4)
      print("Third derivative:", derivatives[3].shape) # shape (2, 2, 2, 3, 4)
    """
    def flatten_tuple(nested_tuple):
        flat_tuple = ()
        for item in nested_tuple:
            if isinstance(item, tuple):
                flat_tuple += flatten_tuple(item)
            else:
                flat_tuple += (item,)
        return flat_tuple

    def jacfwd_upto_n_one_vector_arg_tmp(f, x, n):
        def single_component_derivative(i, f, n):
            # Initialize tangent_directions with zeros like x
            tangent_directions = jnp.zeros_like(x)
            # Set the i-th component to 1
            tangent_directions = tangent_directions.at[i].set(1.0)

            if n == 1:
                # Compute the jvp for this direction
                values = jax.jvp(f, (x,), (tangent_directions,))
                return (values[0],) + (values[-1],)
            else:
                df = lambda y: jacfwd_upto_n_one_vector_arg_tmp(f, y, n-1)
                values = jax.jvp(df, (x,), (tangent_directions,))
                return (values[0][0],) + (values[-1],)

        if n == 0:
            return (f(x),)
        elif n >= 1:
            # Vectorize over each component of x
            jvp_values = jax.vmap(single_component_derivative, in_axes=(0, None, None))(jnp.arange(x.shape[0]), f, n)
            return (jvp_values[0][0],) + (jvp_values[1],)
        else:
            raise ValueError("n must be a non-negative integer.")
    
    return flatten_tuple(jacfwd_upto_n_one_vector_arg_tmp(fun, x, n))



# Copyright (C) 2024 Tobias Bode
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

def test_example_coupled_neumann():

    import jax
    from jax import config
    import jax.numpy as jnp
    import flax

    from autopdex import seeder, geometry, solver, solution_structures, utility, models

    config.update("jax_enable_x64", True)


    # Positive smooth distance functions that select the specific boundary segments
    psdf_bottom = lambda x: geometry.psdf_trimmed_line(x, jnp.stack([0.,0.]), jnp.stack([1.,0.]))
    psdf_right = lambda x: geometry.psdf_trimmed_line(x, jnp.stack([1.,0.]), jnp.stack([1.,1.]))
    psdf_top = lambda x: geometry.psdf_trimmed_line(x, jnp.stack([1.,1.]), jnp.stack([0.,1.]))
    psdf_left = lambda x: geometry.psdf_trimmed_line(x, jnp.stack([0.,1.]), jnp.stack([0.,0.]))
    psdf_cutout = lambda x: geometry.psdf_nd_sphere_cutout(x, jnp.stack([0.5,0.5]), 0.25)
    psdf_rect = lambda x: geometry.psdf_parallelogram(x, jnp.stack([0.,0.]), jnp.stack([0.,1.]), jnp.stack([1.,0.]))
    psdfs = [psdf_bottom, psdf_right, psdf_top, psdf_left, psdf_cutout]

    # Positive smooth distance funcion of full domain (used for seeding nodes and integration points)
    psdf_domain = lambda x: geometry.r_equivalence(psdf_rect(x), psdf_cutout(x))

    # Boundary condition functions
    def homogeneous(x): # Bottom, right and top boundaries
        return 0
    def bc_0(x): # Left boundary
        return jax.lax.sin(jnp.pi * x[0])
    def neumann_at_circ(x): # Normal derivative at circular cutout for a specific analytic solution, may be used for convergence study
        return (2. * jnp.pi / jnp.sinh(jnp.pi)) * ((-1.+2.*x[1])*jnp.cosh(jnp.pi-jnp.pi*x[1]) + (1.-2.*x[0])*jnp.cos(jnp.pi*x[0])*jnp.sinh(jnp.pi-jnp.pi*x[1]))

    # Field selectors for boundary conditions
    def dirichlet(x):
        return jnp.asarray([1., 0., 0.])
    def coupled_neumann(x):
        centroid = jnp.asarray([0.5,0.5])
        dxc = x - centroid
        normal = - dxc / jnp.sqrt(jnp.dot(dxc, dxc))
        return jnp.concatenate([jnp.asarray([0.]), normal])

    # Assignment of boundary conditions
    dirichlet_conditions = [bc_0, homogeneous, homogeneous, homogeneous, neumann_at_circ]
    a_i = [dirichlet, dirichlet, dirichlet, dirichlet, coupled_neumann]

    ### Transfinite interpolation of boundary segments
    b_coeff = lambda x: jnp.asarray([geometry.transfinite_interpolation(x, psdfs, a_i)])
    b_cond = lambda x: jnp.asarray([geometry.transfinite_interpolation(x, psdfs, dirichlet_conditions)])
    psdf = lambda x: jnp.asarray([psdf_domain(x)])


    ### Initialization and generation of nodes and integration points
    # Regular node placement
    nod_spacing = 0.1
    (x_nodes, n_nodes, region_size) = seeder.regular_in_psdf(psdf_domain, [0.,0.], [1.,1.], nod_spacing)

    # Integration points
    # # Regular spacing
    # int_spacing = nod_spacing
    # (x_int, w_int, n_int) = seeder.gauss_points_in_psdf(psdf_domain, [0.,0.], [1.,1.], int_spacing, order=2)
    # Collocation
    x_int = x_nodes
    n_int = n_nodes
    w_int = (region_size / n_int) * jnp.ones(n_int)
    print("Estimated region size: ", w_int.sum())
    assert jnp.isclose(w_int.sum(), 0.49586776859504134), "Region size wrong"

    # Support radius for moving least square ansatz
    support_radius = 2.01 * nod_spacing

    # Definition of pde
    pde = models.poisson_fos(support_radius)

    # Neighborhood search
    (num_neighbors, max_neighbors, min_neighbors, neighbor_list) = utility.search_neighborhood(x_nodes, x_int, support_radius)

    ### Settings
    n_fields = 3
    dofs0 = jnp.zeros((n_nodes, n_fields))
    static_settings = flax.core.FrozenDict({
    'solution space': ('mls',),
    'shape function mode': 'compiled',
    'number of fields': (n_fields,),
    'assembling mode': ('sparse',),
    'maximal number of neighbors': (max_neighbors,),
    'order of basis functions': (1,),
    'weight function type': ('gaussian',),
    'variational scheme': ('least square pde loss',),
    'model': (pde,),
    'solution structure': ('first order set',),
    'boundary coefficients': (b_coeff,),
    'boundary conditions': (b_cond,),
    'psdf': (psdf,),
    'solver type': 'linear',
    'solver backend': 'scipy',
    'solver': 'lapack',
    'verbose': 1,
    'connectivity': (utility.jnp_to_tuple(neighbor_list),),
    })
    settings = {
    'beta': (3,),
    'node coordinates': x_nodes,
    'integration coordinates': (x_int,),
    'integration weights': (w_int,),
    'support radius': (support_radius,),
    }

    ### Optional: Precompute shape functions and pass them to e.g. static_settings...
    settings = solution_structures.precompile(dofs0, settings, static_settings)


    ### Call solver
    dofs, _ = solver.solver(dofs0, settings, static_settings, tol=1e-8)

    check = dofs.flatten().sum()
    # print(check)
    assert jnp.isclose(check, -22.904093385453102), "Solution wrong"



# Copyright (C) 2024 Tobias Bode
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

"""
This example demonstrates the solution of Cook's membrane problem using custom elements from 
the `autopdex.models` library and an exemplary external AceGen-generated C routine.

The main steps include:
    1. Definition of geometry and boundary conditions.
    2. Mesh generation using `pygmsh` and conversion to JAX arrays.
    3. Selection of nodes for boundary conditions and setup of Dirichlet conditions.
    4. Definition of material properties for the neo-Hookean model.
    5. Optional use of an AceGen-generated shared library for custom element computation.
    6. Element definition for hyperelastic steady-state weak form.
    7. Element definition for imposition of Neumann boundary conditions.
    8. Solver settings and configuration using `flax.core.FrozenDict`.
    9. Adaptive load stepping solver for nonlinear analysis.
    10. Export results in VTK format for ParaView.
"""

if __name__ == "__main__":
    import time
    import sys
    import ctypes
    import os

    import jax
    from jax import config
    import numpy as np
    import jax.numpy as jnp
    import flax
    import pygmsh
    import meshio

    from autopdex import seeder, geometry, solver, solution_structures, plotter, utility, models, spaces, assembler

    config.update("jax_enable_x64", True)
    config.update("jax_compilation_cache_dir", './cache')


    ### Definition of geometry and boundary conditions and mesh
    order = 2
    pts = [[0.,0.], [48.,44.], [48.,60.], [0.,44.]]
    with pygmsh.occ.Geometry() as geom:
        region = geom.add_polygon(pts, mesh_size=1.0)
        # region = geom.add_polygon(pts, mesh_size=0.3)
        geom.set_recombined_surfaces([region.surface])
        mesh = geom.generate_mesh(order=order)
        print("Mesh generation finished.")

    # Import mesh
    n_dim = 2
    x_nodes = jnp.asarray(mesh.points[:,:n_dim])
    n_nodes = x_nodes.shape[0]
    elements = jnp.asarray([ v for k, v in mesh.cells_dict.items() if 'quad' in k])[0]
    surface_elements = jnp.asarray([ v for k, v in mesh.cells_dict.items() if 'line' in k])[0]
    print("Number of elements: ", elements.shape[0])
    print("Number of unknowns: ", n_nodes * n_dim)

    # Selection of nodes for boundary condition
    dirichlet_nodes = geometry.in_planes(x_nodes, pts[0], [1.,0.])
    dirichlet_dofs = utility.dof_select(dirichlet_nodes, jnp.asarray([True, True]))
    dirichlet_conditions = jnp.zeros_like(dirichlet_dofs, dtype=jnp.float_)

    # Import surface mesh for inhomogeneous Neumann conditions
    neumann_selection = geometry.select_elements_in_plane(x_nodes, surface_elements, pts[1], [1.,0.])
    neumann_elements= surface_elements[neumann_selection]

    # Definition of material parameters
    lam = 100.
    mu = 40.
    Em = mu * (3 * lam + 2 * mu) / (lam + mu)
    nu = lam / (2 * (lam + mu))


    ### Select element
    element = 'built-in'
    # element = 'acegen'

    if element == 'built-in':
        ### Built-in AutoPDEx element
        weak_form_fun_1 = models.hyperelastic_steady_state_weak(models.neo_hooke, 
                                                            lambda x: Em,
                                                            lambda x: nu, 
                                                            'plain strain')
        user_elem_1 = models.isoparametric_domain_element_galerkin(weak_form_fun_1, 
                                                                spaces.fem_iso_line_quad_brick,
                                                                *seeder.gauss_legendre_nd(dimension = 2, order = 2 * order))
    elif element == 'acegen':
        ### Load the shared library with AceGen generated element 
        # The example element can be generated with the notebook 'external_user_elem_generation.nb' 
        # using AceGen(http://symech.fgg.uni-lj.si/Download.htm) and Mathematica
        # You have to add extern "C" in the beginning of the document, uncomment #include "sms.h" 
        # and move the decleration of the vector v[...] into the function
        # Then, you can compile it e.g. with
        # g++ -shared -o user_elem.dll user_elem.cpp -O2 -s -fPIC
        # and adjust the following pat
        path_to_code = 'examples/cooks_membrane/user_elem.dll'
        lib = ctypes.CDLL(path_to_code)

        # Define the ctypes types corresponding to the function signature
        DoublePtr = ctypes.POINTER(ctypes.c_double)
        Double2DArray = np.ctypeslib.ndpointer(dtype=np.float64, ndim=2, flags='C_CONTIGUOUS')
        Double1DArray = np.ctypeslib.ndpointer(dtype=np.float64, ndim=1, flags='C_CONTIGUOUS')

        # Define the function prototype
        lib.user_elem.argtypes = [
            Double2DArray,  # double fI[9][2]
            Double2DArray,  # double xI[9][2]
            DoublePtr,      # double (*Em)
            DoublePtr,      # double (*nu)
            Double1DArray,  # residual
            Double2DArray   # tangent
        ]
        lib.user_elem.restype = None

        # Define a Python callback function
        def my_c_function_callback(fI, xI, Em, nu):
            fI_np = np.ascontiguousarray(np.asarray(fI))
            xI_np = np.ascontiguousarray(np.asarray(xI))

            residual = np.zeros((18,), dtype=np.float64)
            tangent = np.zeros((18, 18), dtype=np.float64)

            # Convert scalars to ctypes pointers
            Em_ct = ctypes.byref(ctypes.c_double(Em))
            nu_ct = ctypes.byref(ctypes.c_double(nu))

            # Call the C function
            lib.user_elem(fI_np, xI_np, Em_ct, nu_ct, residual, tangent)
            return residual, tangent

        def user_elem_1(fI, xI, elem_number, settings, static_settings, mode, set):
            result_shape_dtype = (
                jax.ShapeDtypeStruct(shape=(18,), dtype=np.float64),
                jax.ShapeDtypeStruct(shape=(18, 18), dtype=np.float64)
            )
            arguments = (fI, xI, Em, nu)
            
            # Call the pure_callback with the appropriate arguments
            residual, tangent = jax.pure_callback(my_c_function_callback, result_shape_dtype, *arguments)

            if mode == 'residual':
                return residual
            elif mode == 'tangent':
                return tangent
            else:
                raise ValueError("User elem mode has to be 'residual' or 'tangent'.")
    else:
        assert False, "Element not set. Use 'acegen' or 'built-in'."

    # Tractions suitable for adaptive load stepping
    q_0 = -1.5e+1
    # traction_fun = lambda x: jnp.asarray([0., q_0])
    traction_fun = lambda x, settings: jnp.asarray([0., settings['load multiplier']])
    weak_form_fun_2 = models.neumann_weak(traction_fun)
    user_elem_2 = models.isoparametric_surface_element_galerkin(weak_form_fun_2, 
                                                            spaces.fem_iso_line_quad_brick,
                                                            *seeder.gauss_legendre_nd(dimension = 1, order = 2 * order),
                                                            tangent_contributions=False)

    ### Settings
    n_fields = 2
    dofs0 = jnp.zeros((n_nodes, n_fields))
    static_settings = flax.core.FrozenDict({
    'number of fields': (n_fields, n_fields),
    'assembling mode': ('user element', 'user element'),
    'solution structure': ('nodal imposition', 'nodal imposition'),
    'model': (user_elem_1, user_elem_2),
    'solver type': 'newton',
    'solver backend': 'pardiso',
    'solver': 'lu',
    'verbose': 1,
    'dirichlet dofs': utility.jnp_to_tuple(dirichlet_dofs),
    'connectivity': (utility.jnp_to_tuple(elements), utility.jnp_to_tuple(neumann_elements)),
    })

    settings = {
    'load multiplier': q_0,
    'node coordinates': x_nodes,
    'dirichlet conditions': dirichlet_conditions,
    }


    # ### Timing comparison for computing tangent contributions.
    # start = time.time()
    # tangent_contributions = assembler.user_element_assemble_tangent(dofs0, settings, static_settings, 0)
    # print("Compile time: ", time.time() - start)
    # start = time.time()
    # for i in range(10):
    #     tangent_contributions = assembler.user_element_assemble_tangent(dofs0, settings, static_settings, 0)
    # print("Time needed: ", time.time() - start)
    # sys.exit()


    # Adaptive load stepping
    def multiplier_settings(settings, multiplier):
        settings['load multiplier'] = multiplier * q_0
        return settings
    
    dofs = solver.adaptive_load_stepping(dofs = dofs0, 
                                        settings = settings, 
                                        static_settings = static_settings,
                                        multiplier_settings = multiplier_settings,
                                        path_dependent = False)[0]

    ### Postprocessing
    points = mesh.points
    cells = jnp.asarray([ v for k, v in mesh.cells_dict.items() if 'quad' in k])[0]
    mesh = meshio.Mesh(
        points,
        {'quad': cells[:,:4]},
        point_data={
            "u": dofs,
        },
    )
    mesh.write("cooks_membrane.vtk")



# Copyright (C) 2024 Tobias Bode
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

"""
Dirichlet problem of the Laplace equation in 2d: [0,1]x[0,1] with cutout circle (midpoint=[0.5,0.5], radius=0.25)
Demonstration of quasi-random sampling in regions defined by signed distance functions
"""

if __name__ == "__main__":

  import jax
  from jax import lax, config
  import jax.numpy as jnp
  import flax

  from autopdex import seeder, geometry, solver, solution_structures, plotter, utility, models

  config.update("jax_enable_x64", True)
  config.update("jax_compilation_cache_dir", './cache')


  # Analytic solution for calculation of L2 error norm
  def analytic_solution(x):
    return jnp.stack([lax.sinh(jnp.pi * x[0] / 2) * lax.sin(jnp.pi * x[1] / 2)])

  # Positive smooth distance functions that select the specific boundary segments
  psdf_bottom = lambda x: geometry.psdf_trimmed_line(x, [0.,0.], [1.,0.])
  psdf_right = lambda x: geometry.psdf_trimmed_line(x, [1.,0.], [1.,1.])
  psdf_top = lambda x: geometry.psdf_trimmed_line(x, [1.,1.], [0.,1.])
  psdf_left = lambda x: geometry.psdf_trimmed_line(x, [0.,1.], [0.,0.])
  psdf_cutout = lambda x: geometry.psdf_nd_sphere_cutout(x, [0.5,0.5], 0.25)
  psdf_rect = lambda x: geometry.psdf_parallelogram(x, [0.,0.], [0.,1.], [1.,0.])
  psdfs = [psdf_bottom, psdf_right, psdf_top, psdf_left, psdf_cutout]


  # Positive smooth distance funcion of full domain (used for seeding nodes and integration points)
  psdf_domain = lambda x: geometry.r_equivalence(psdf_rect(x), psdf_cutout(x))

  # Boundary condition functions
  def homogeneous(x): # Bottom, right and top boundaries
    return 0
  def bc_0(x): # Left boundary
    return jax.lax.sin(jnp.pi * x[0])


  # Assignment of boundary conditions
  dirichlet_conditions = [bc_0, homogeneous, homogeneous, homogeneous, lambda x: 1.0]


  ### Transfinite interpolation of boundary segments
  b_cond = lambda x: geometry.transfinite_interpolation(x, psdfs, dirichlet_conditions)
  psdf = psdf_domain

  ### Initialization and generation of nodes and integration points
  # Quasi-random node placement
  n_nodes = 10000
  # (x_nodes, n_nodes, region_size) = seeder.quasi_random_in_psdf(psdf_domain, [0.,0.], [1.,1.], n_nodes, 'hammersley')
  (x_nodes, n_nodes, region_size) = seeder.quasi_random_in_psdf(psdf_domain, [0.,0.], [1.,1.], n_nodes, 'halton')

  # Integration points
  # Collocation, Note: this can in some cases cause bad conditioning, unphysical low-energy modes and divergence. 
  # If so, use more integration points, e.g. Gauss integration on a background grid (seeder.gauss_points_in_psdf)
  x_int = x_nodes
  n_int = n_nodes

  # Support radius for moving least square ansatz
  support_radius = 3.01 * lax.sqrt(region_size/n_nodes)

  # Integration point weights
  w_int = (region_size / n_int) * jnp.ones(n_int)

  # Definition of pde
  pde = models.poisson(source_fun=None)


  ### Neighborhood search
  (num_neighbors, max_neighbors, min_neighbors, neighbor_list) = utility.search_neighborhood(x_nodes, x_int, support_radius)


  ### Settings
  n_fields = 1
  dofs0 = jnp.zeros((n_nodes, n_fields))
  static_settings = flax.core.FrozenDict({
    'solution space': ('mls',),
    'shape function mode': 'compiled',
    'number of fields': (n_fields,),
    'assembling mode': ('sparse',),
    'maximal number of neighbors': (max_neighbors,),
    'order of basis functions': (2,),
    'weight function type': ('gaussian',),
    'variational scheme': ('least square pde loss',),
    # 'variational scheme': ('strong form galerkin',),
    'model': (pde,),
    'solution structure': ('dirichlet',),
    'boundary conditions': (b_cond,),
    'psdf': (psdf,),
    'solver type': 'linear',
    'solver backend': 'pardiso',
    'solver': 'lu',
    # 'solver backend': 'jax',
    # 'solver': 'bicgstab',
    # 'type of preconditioner': 'jacobi',
    # 'hvp type': 'linearize',
    'verbose': 1,
    'connectivity': (utility.jnp_to_tuple(neighbor_list),),
  })
  settings = {
    'beta': (3,),
    'node coordinates': x_nodes,
    'integration coordinates': (x_int,),
    'integration weights': (w_int,),
    'support radius': (support_radius,),
  }

  # Precompute solution structure
  settings = solution_structures.precompile(dofs0, settings, static_settings)

  ### Call solver
  dofs, _ = solver.solver(dofs0, settings, static_settings, tol=1e-8, newton_tol=1e-8)

  ### Preperation of postprocessing data
  local_dofs = dofs[neighbor_list]
  post_fun_vj = jax.jit(jax.vmap(solution_structures.solution_structure, (0,0,0,None,None,None)), static_argnames=['static_settings', 'set'])
  data = post_fun_vj(x_int, jnp.arange(n_int), local_dofs, settings, static_settings, 0)

  # Plotting with pyvista
  plotter.pv_plot(x_int, data[:,0], export_vtk=False, show=True)

  # Export to Paraview
  import meshio
  import numpy as np
  points = x_int
  cells = [("vertex", np.array([[i] for i in range(len(points))]))]
  mesh = meshio.Mesh(
      points,
      cells,
      point_data={
          "temperature": data[:,0],
      },
  )
  mesh.write("laplace.vtk")
